#include <stdio.h>
#include "mstring.h"
#include "complz77.h"
#include "math.h"

/*
#ifdef	WIN32
	#ifndef _CONSOLE
		#include <windows.h>
	#endif
#endif
*/


CompressLZ77::CompressLZ77()
:Compress()
{
	typeChar = '^';
}

CompressLZ77::~CompressLZ77()
{
	//nothing to destruct
}

/*
 * This is the compression routine.  It has to first load up the look
 * ahead buffer, then go into the main compression loop.  The main loop
 * decides whether to output a single character or an index/length
 * token that defines a phrase.  Once the character or phrase has been
 * sent out, another loop has to run.  The second loop reads in new
 * characters, deletes the strings that are overwritten by the new
 * character, then adds the strings that are created by the new
 * character.
 */
void CompressLZ77::CompressFile( String &input, String &output )
{
	bit_file	*in;
	bit_file	*out;
    int			i;
    int			c;
    int			look_ahead_bytes;
    int			current_position;
    int			replace_count;
    int			match_length;
    int			match_position;


	in  = OpenInputBitFile( input );
	out = OpenOutputBitFile( output );

    current_position = 1;
    for ( i = 0 ; i < LOOK_AHEAD_SIZE ; i++ ) {
        if ( ( c = InputStd( in ) ) == EOF )
            break;
        window[ current_position + i ] = (unsigned char) c;
    }
    look_ahead_bytes = i;
    InitTree( current_position );
    match_length = 0;
    match_position = 0;
    while ( look_ahead_bytes > 0 ) {
        if ( match_length > look_ahead_bytes )
            match_length = look_ahead_bytes;
        if ( match_length <= BREAK_EVEN ) {
            replace_count = 1;
            OutputBit( *out, 1 );
            OutputBits( *out,(unsigned long) window[ current_position ], 8 );
        } 
		else {
            OutputBit( *out, 0 );
            OutputBits( *out,(unsigned long) match_position, INDEX_BIT_COUNT );
            OutputBits( *out,(unsigned long) ( match_length - ( BREAK_EVEN + 1 ) ), LENGTH_BIT_COUNT );
            replace_count = match_length;
        }
        for ( i = 0 ; i < replace_count ; i++ ) {
            DeleteString( MOD_WINDOW( current_position + LOOK_AHEAD_SIZE ) );
            if ( ( c = InputStd( in ) ) == EOF )
                look_ahead_bytes--;
            else
                window[ MOD_WINDOW( current_position + LOOK_AHEAD_SIZE ) ]
                        = (unsigned char) c;
            current_position = MOD_WINDOW( current_position + 1 );
            if ( look_ahead_bytes )
                match_length = AddString( current_position, &match_position );
        }
    };
    OutputBit( *out, 0 );
    OutputBits( *out, (unsigned long) END_STREAM, INDEX_BIT_COUNT );

	CloseOutputBitFile( out );
	CloseInputBitFile( in );
}

/*
 * This is the expansion routine for the LZSS algorithm.  All it has
 * to do is read in flag bits, decide whether to read in a character or
 * a index/length pair, and take the appropriate action.
*/
void CompressLZ77::ExpandFile( String &input, String &output )
{
	bit_file    *in;
	bit_file    *out;
    int			i;
    int			current_position;
    int			c;
    int			match_length;
    int			match_position;


	in  = OpenInputBitFile( input );
	out = OpenOutputBitFile( output );

    current_position = 1;
    for ( ; ; ) {
        if ( InputBit( *in ) ) {
            c = (int) InputBits( *in, 8 );
            OutputStd( c, out );
            window[ current_position ] = (unsigned char) c;
            current_position = MOD_WINDOW( current_position + 1 );
        } else {
            match_position = (int) InputBits( *in, INDEX_BIT_COUNT );
            if ( match_position == END_STREAM )
                break;
            match_length = (int) InputBits( *in, LENGTH_BIT_COUNT );
            match_length += BREAK_EVEN;
            for ( i = 0 ; i <= match_length ; i++ ) {
                c = window[ MOD_WINDOW( match_position + i ) ];
                OutputStd( c, out );
                window[ current_position ] = (unsigned char) c;
                current_position = MOD_WINDOW( current_position + 1 );
            }
        }
    }

	CloseOutputBitFile( out );
	CloseInputBitFile( in );
}

/*
 * To make the tree usable, a single phrase has to be
 * added to the tree so it has a root node.
*/
void CompressLZ77::InitTree( int r )
{
    tree[ TREE_ROOT ].larger_child = r;
    tree[ r ].parent = TREE_ROOT;
    tree[ r ].larger_child = UN_USED;
    tree[ r ].smaller_child = UN_USED;
}

/*
 * This routine is used when a node is being deleted.  The link to
 * its descendant is broken by pulling the descendant in to overlay
 * the existing link.
 */
void CompressLZ77::ContractNode( int old_node, int new_node )
{
    tree[ new_node ].parent = tree[ old_node ].parent;
    if ( tree[ tree[ old_node ].parent ].larger_child == old_node )
        tree[ tree[ old_node ].parent ].larger_child = new_node;
    else
        tree[ tree[ old_node ].parent ].smaller_child = new_node;
    tree[ old_node ].parent = UN_USED;
}

/*
 * This routine is also used when a node is being deleted.  However,
 * in this case, it is being replaced by a node that was not previously
 * in the tree.
 */
void CompressLZ77::ReplaceNode( int old_node, int new_node )
{
    int parent;

    parent = tree[ old_node ].parent;
    if ( tree[ parent ].smaller_child == old_node )
        tree[ parent ].smaller_child = new_node;
    else
        tree[ parent ].larger_child = new_node;
    tree[ new_node ] = tree[ old_node ];
    tree[ tree[ new_node ].smaller_child ].parent = new_node;
    tree[ tree[ new_node ].larger_child ].parent = new_node;
    tree[ old_node ].parent = UN_USED;
}

/*
 * This routine is used to find the next smallest node after the node
 * argument.  It assumes that the node has a smaller child.  We find
 * the next smallest child by going to the smaller_child node, then
 * going to the end of the larger_child descendant chain.
*/
int CompressLZ77::FindNextNode( int node )
{
    int next;

    next = tree[ node ].smaller_child;
    while ( tree[ next ].larger_child != UN_USED )
        next = tree[ next ].larger_child;
    return( next );
}

/*
 * This routine performs the classic binary tree deletion algorithm.
 * If the node to be deleted has a null link in either direction, we
 * just pull the non-null link up one to replace the existing link.
 * If both links exist, we instead delete the next link in order, which
 * is guaranteed to have a null link, then replace the node to be deleted
 * with the next link.
 */
void CompressLZ77::DeleteString( int p )
{
    int  replacement;

    if ( tree[ p ].parent == UN_USED )
        return;
    if ( tree[ p ].larger_child == UN_USED )
        ContractNode( p, tree[ p ].smaller_child );
    else if ( tree[ p ].smaller_child == UN_USED )
        ContractNode( p, tree[ p ].larger_child );
    else {
        replacement = FindNextNode( p );
        DeleteString( replacement );
        ReplaceNode( p, replacement );
    }
}

/*
 * This where most of the work done by the encoder takes place.  This
 * routine is responsible for adding the new node to the binary tree.
 * It also has to find the best match among all the existing nodes in
 * the tree, and return that to the calling routine.  To make matters
 * even more complicated, if the new_node has a duplicate in the tree,
 * the old_node is deleted, for reasons of efficiency.
 */
int CompressLZ77::AddString( int new_node, int *match_position )
{
    int i;
    int test_node;
    int delta;
    int match_length;
    int *child;

    if ( new_node == END_STREAM )
        return( 0 );
    test_node = tree[ TREE_ROOT ].larger_child;
    match_length = 0;
    for ( ; ; ) {
        for ( i = 0 ; i < LOOK_AHEAD_SIZE ; i++ ) {
            delta = window[ MOD_WINDOW( new_node + i ) ] -
                    window[ MOD_WINDOW( test_node + i ) ];
            if ( delta != 0 )
                break;
        }
        if ( i >= match_length ) {
            match_length = i;
            *match_position = test_node;
            if ( match_length >= LOOK_AHEAD_SIZE ) {
                ReplaceNode( test_node, new_node );
                return( match_length );
            }
        }
        if ( delta >= 0 )
            child = &tree[ test_node ].larger_child;
        else
            child = &tree[ test_node ].smaller_child;
        if ( *child == UN_USED ) {
            *child = new_node;
            tree[ new_node ].parent = test_node;
            tree[ new_node ].larger_child = UN_USED;
            tree[ new_node ].smaller_child = UN_USED;
            return( match_length );
        }
        test_node = *child;
    }
}

