/** =================================================================\n
 *
 * MathLib\n
 * @file Matrix.cpp
 *
 * Primary Author: Howard Zhou (howardz@cc.gatech.edu)\n
 *
 * (C) Copyright Howard Zhou 2003\n
 * (C) Copyright Georgia Institute of Technology, 2003\n
 *
 * ==================================================================\n
 * Basic matrix lab
 *
 * @version 1.1\n
 *     Major organization change, moved most of function defintion to\n
 *     the .cpp file to make class declaration more clear.
 * @date 2003-05-14
 * @author howardz
 *
 * @version 1.0\n
 *     Initial Revision
 * @date 2003-02-23
 * @author howardz
 *
 *
 * @todo Implement det() and invert() 
 * @bug
 * @warning 
 *
 * ===================================================================\n
 */

#include "Matrix.h"

#include <stdio.h>
#include <math.h>
#include <assert.h>
#include <string.h>
#include <iostream>
#include <iomanip>


#define SWAP(x, y) (x)=(x)+(y); (y)=(x)-(y); (x)=(x)-(y)

/// @namespace using standard namespace
using namespace std;

// ======================================================
// Actual function definition
// ======================================================


// ------------------------------------------------------
// Accessors and Mutators
// ------------------------------------------------------

/**
 * Return the determinant of the matrix
 * @return double: the determinant of the matrix
 */
double CMatrix::det(void) const
{
	return	1;
}

/**
 * Inverts the current matrix.
 * @return reference to this
 */
CMatrix& CMatrix::invert (void) 
{
	return(*this);
}

/**
 * Transposes (swithces rows and columns in) the current matrix.
 * @return 
 */
CMatrix& CMatrix::transpose (void) 
{
	int i, j;
	if(isSquare())
	{
		for(i = 0; i< _rows; i++)
			for(j=0; j< _cols; j++)
				if(i < j)
					swap(EL(_m,i,j), EL(_m, j, i));
	}
	else
	{
		CMatrix mat = CMatrix(*this);
		SWAP(_rows, _cols);
		for(i = 0; i< _rows; i++)
			for(j=0; j< _cols; j++)
					EL(_m,i,j) = mat[j][i];				
	}

	return(*this);
}


// ------------------------------------------------------
// Overloaded Operators 
// ------------------------------------------------------

/**
 * operator *
 * Multiply the refvector to the right of the matrix.
 * @param refvector reference to a CVector
 * @return a CVector that is the product of the current matrix and refvector.
 */
CVector CMatrix::operator* (const CVector& refvector) const 
{
	assert(isMultiplicable(refvector));
		
	CVector vec(_rows);

	for(int i = 0; i < _rows; i++)
		for(int j=0;j<_cols; j++)
			vec[i] += EL(_m, i, j) * refvector[j];
	
	return CVector(vec);
}

/**
 * operator *
 * Multiply the refmatrix to the right of the matrix.
 * @param refmatrix reference to a CMatrix
 * @return a matrix that is the product of the current matrix and refmatrix.
 */
CMatrix CMatrix::operator* (const CMatrix& refmatrix) const 
{
	assert(isMultiplicable(refmatrix));	// _cols == refmatrix._rows

	int rcols = refmatrix._cols;
	CMatrix mat(_rows, rcols);
	mat = 0;	// initialize every entry in this matrix to 0

	for(int i = 0; i < _rows; i++)
		for(int j = 0;j < rcols; j++)
			for(int k=0; k < _cols; k++)
				mat[i][j] += EL(_m, i, k) * refmatrix[k][j];

	return CMatrix(mat);
}

/**
 * operator *=
 * Multiplies the current matrix by refmatrix on the right
 * @param refmatrix reference to a matrix
 * @return reference to this 
 */
CMatrix& CMatrix::operator*= (const CMatrix& refmatrix) 
{
	assert(isCompatible(refmatrix));	// _cols == refmatrix._rows
	
	CMatrix mat(refmatrix);
	mat = 0;	// initialize every entry in this matrix to 0

	for(int i = 0; i < _rows; i++)
		for(int j = 0;j < _cols; j++)
			for(int k=0; k < _cols; k++)
				mat[i][j] += EL(_m, i, k) * refmatrix[k][j];
	*this = mat;
	return (*this);
}


// ------------------------------------------------------
// Output or Test
// ------------------------------------------------------

/**
 * Stand along display
 */
void CMatrix::display() const
{
	assert(isValid());
	printf("\n");
	for(int i=0; i< _rows; i++)
	{
		printf("| ");
		for(int j= 0; j< _cols; j++)
			printf("%f ", _m[i*_cols + j]);
		printf("|\n");
	}
}

/**
 * operator <<:
 * Display matrix through the ostream
 * @param os reference to the ostream object
 * @param refmatrix reference to a matrix 
 * @return ostream with the new stream of mat contents attached
 */
ostream& operator<< (ostream& os, CMatrix& refmatrix)
{
	assert(refmatrix.isValid());
	for(int i=0; i< refmatrix._rows; i++)
	{
		os << "| ";
		for(int j= 0; j<refmatrix._cols; j++)
		{	
			using namespace std;
			os << setprecision(5);
			os << setw(10);
			os << setiosflags (ios_base::fixed | ios_base::right);
			os << refmatrix._m[i*refmatrix._cols + j] << "  ";
		}
		os << "|\n";
	}
	return os;
}

/**
 * General test function for CMatrix class
 * put anything you want to test here.
 */
void CMatrix::test()
{
	cout << "\n\nCMatrix Test\n=========================\n";

	CMatrix m1;
	m1.reSize(3, 3);
	m1.display();
	cout << m1 << endl;

	m1[2][1] = 24;
	m1[0][2] = 15;
	m1[1][0] = -4;
	cout << m1 << endl;
	
	CMatrix m2 = m1.transposed();
	cout << m2 << endl;

	cout << "equal? " << (m1 == m2) << endl;
	cout << "not equal? " << (m1 != m2) << endl;

	cout << "is symmetric? " << m1.isSymmetric() << endl;

	CMatrix m3;
	m3.reSize(3, 3);
	m3.identity();
	cout << m3 << endl; 

	CMatrix m4 = m1*m3;
	cout << m4 << endl;

	m4 = m1*m1;
	cout << m4 << endl;


	CMatrix m5;
	m5.reSize(3, 5);
	m5 = 2;
	CMatrix m6 = m3*m5;
	m6[0][4] = 0.5;
	m6[2][1] = 16;
	m6[1][1] = 1;
	cout << m6 << endl;

	m6.transpose();
	cout << m6 << endl;

	CMatrix m7;
	m7.reSize(1,4);
	m7[0][0] = 1;
	m7[0][1] = 2;
	m7[0][2] = 3;
	m7[0][3] = 4;
	cout << m7 << endl;

	m7.transpose();
	cout << m7 << endl;

	CMatrix m8 = CMatrix(2, 2);
	m8 = 1;
	CMatrix m9 = CMatrix(2, 2);
	m9 = 2;
	CMatrix m10 = CMatrix(2, 2);
	m10 = 3;

	CMatrix m11 = m8;
	CMatrix m12 = m8;
	CMatrix m13 = CMatrix(2, 2);
	m13 = 0;

	m8 += m9 * m10;

	cout << m8 << endl;

	m11 = m11 + m9*m10;
	cout << m11 << endl;

	m13 = m9*m10;
	cout << m13 << endl;

	m12 = m12+m13;
	cout << m12 << endl;

	CMatrix m14, m15, m16, m17, m18;
	m15.reSize(3, 3);
	m16.reSize(3, 3);
	m14.reSize(3, 3);
	m17.reSize(3, 3);
	m18.reSize(3, 3);
	/*
	CVector v1(3, 2);
	CVector v2 = m3* v1;
	cout << v2 << " = " << m3 << " * " << v1 << endl;
	*/

	//CVector v1(4, 1);
	//v1.display();
	//CVector v2(4, 2);
	//v2.display();

	//v1 = v2;
	//v1.display();
	//v2 = 3;
	//v2[2] = -2;
	//v2.display();
	//v1 = -v2;
	//v1.display();
	//v2.display();
	//printf("length %d, magnitude %f\n", v2.length(), v2.magnitude());
	//v2.normalize();
	//v2.display();

	//CVector v3(4, 5);
	//v3[2] = 0;
	//v3[3] = 0;
	//v3[4] = 12;
	//v3.display();
	//printf("length %d, magnitude %f\n", v3.length(), v3.magnitude());
	//v3.normalize();
	//printf("length %d, magnitude %f\n", v3.length(), v3.magnitude());
	//v3.display();

	//v3[1] = 0;
	//v3[4] = 0;
	//v3[3] = 1;
	//printf("length %d, magnitude %f\n", v3.length(), v3.magnitude());
	//v3.normalize();
	//printf("length %d, magnitude %f\n", v3.length(), v3.magnitude());
	//v3.display();		
	//
	//v3[1] = 5;
	//v3[2] = 0;
	//v3[3] = 6;
	//v3[4] = 12;


	//printf("+, +=, -, -=, *, *=, /, /=\n");
	//v1= CVector(4, 1);
	//v2= CVector(4);

	//CVector v4 = v1+v2;
	//v4.display();
	//v4 += v1;
	//v4.display();
	//v4 = v1 - v2;
	//v4.display();
	//v4 -= v1;
	//v4.display ();

	//v3.display();
	//printf("v3 * v3 = %f\n", v3 * v3);
	//v4.display ();
	//v4 = CVector(4, 1);
	//v4.display ();
	//v4 *= 2;
	//v4.display();
	//v4 = v4 - v1;
	//v4 -= v4;
	//v4.display();
	//v2.display();

	//if(v2 == v4)
	//	printf("equal");
	//else
	//	printf("not equal");
}
//@}
