/** =================================================================\n
 *
 * MathLib\n
 * @file Matrix.h
 *
 * Primary Author: Howard Zhou (howardz@cc.gatech.edu)\n
 *
 * (C) Copyright Howard Zhou 2003\n
 * (C) Copyright Georgia Institute of Technology, 2003\n
 *
 * ==================================================================\n
 * Basic matrix lab
 *
 * @version 1.1\n
 *     Major organization change, moved most of function defintion to\n
 *     the .cpp file to make class declaration more clear.
 * @date 2003-05-14
 * @author howardz
 *
 * @version 1.0\n
 *     Initial Revision
 * @date 2003-02-23
 * @author howardz
 *
 *
 * @todo Implement det() and invert() 
 * @bug
 * @warning 
 *
 * ===================================================================\n
 */

#ifndef ___MATRIX__H___
#define ___MATRIX__H___

#include "Vector.h"

#include <stdio.h>
#include <assert.h>

/// @def Fast access to any matrix element through common indexing method
#define EL(x,i,j) ((x)[(i)*_cols+(j)])


/// @namespace using standard namespace
using namespace std;

// ==================================================================================
/**
 * @class CMatrix Matrix.h Matrix.h
 * @brief General Matrix class
 * @ingroup Lib
 *
 * @par requirements
 *
 * @par license
 * This code is absolutely free to use and modify. The code is provided "as is" with
 * no expressed or implied warranty. The author accepts no liability if it causes
 * any damage to your computer, causes your pet to fall ill, increases baldness
 * or makes your car start emitting strange noises when you start it up.
 * This code has no bugs, just undocumented features!
 *
 * @version 1.1\n
 *     Major organization change, moved most of function defintion to the .cpp\n
 *     file to make class declaration more clear.
 * @date 2003-05-14
 * @author howardz
 * 
 * @version 1.0\n
 *     Initial Revision
 * @date 2003-02-23
 * @author howardz
 *
 * @todo Implement det() and invert()
 * @bug none
 * @warning none
 */
// ================================================================================
class CMatrix
{
protected:
	int _rows;      ///< number of rows
	int _cols;      ///< number of columns
	double *_m;     ///< double array to hold data

private:
	// ------------------------------------------------------
	// private helpers
	// ------------------------------------------------------
 	void swap(double &x, double &y);

public:
	// ------------------------------------------------------
	// constructors and destructors
	// ------------------------------------------------------

	/// Default constructor. Creates an uninitialized matrix.
	CMatrix (void){_rows = 0;_cols = 0;_m = 0;};		
	CMatrix (int rows, int cols, double *storage = 0);
	CMatrix (const CMatrix &refmatrix);

	void reSize (int rows, int cols, double *storage = 0);

	~CMatrix (void);	// Destructor

	// ------------------------------------------------------
	// Predicators
	// ------------------------------------------------------
	bool isValid        (void) const;
	bool isSquare       (void) const;
	bool isSymmetric    (void) const;
	bool isIdentity     (void) const;
	bool isCompatible   (const CMatrix& refmatrix) const;
	bool isMultiplicable(const CVector& refvector) const;
	bool isMultiplicable(const CMatrix& refmatrix) const;

	// ------------------------------------------------------
	// Accessors and Mutators
	// ------------------------------------------------------
	int  getRows        (void) const;
	int  getCols        (void) const;
	CMatrix& identity   (void);// make the current matrix an identity matrix
	double   det        (void) const;
	CMatrix& invert     (void);
	CMatrix  inverted   (void) const;
	CMatrix& transpose  (void);
	CMatrix  transposed (void) const;

	// ------------------------------------------------------
	// Overloaded Operators 
	// ------------------------------------------------------
	double*  operator[] (int i) const;
	CMatrix& operator=  (double value); 
	CMatrix& operator=  (const  CMatrix& refmatrix);
	CMatrix  operator-  (void)  const;
	CMatrix  operator+  (const  CMatrix& refmatrix) const;
	CMatrix  operator-  (const  CMatrix& refmatrix) const;
	CMatrix  operator*  (double dbl) const;
	CVector  operator*  (const  CVector& refvector) const;
	CMatrix  operator*  (const  CMatrix& refmatrix) const; 
	CMatrix  operator/  (double dbl) const;
	CMatrix& operator+= (const  CMatrix& refmatrix);
	CMatrix& operator-= (const  CMatrix& refmatrix);
	CMatrix& operator*= (double dbl);
	CMatrix& operator*= (const  CMatrix& refmatrix);
	CMatrix& operator/= (double dbl);
	bool     operator== (const  CMatrix& refmatrix) const;
	bool     operator!= (const  CMatrix& refmatrix) const;
	bool     operator<  (const  CMatrix& refmatrix) const;


	// ------------------------------------------------------
	// Output or Test
	// ------------------------------------------------------
	void     display() const;
	void	    test();

	// ------------------------------------------------------
	// Friend functions
	// ------------------------------------------------------
	friend std::ostream& operator<< (std::ostream& os,	CMatrix& refmatrix);

};


// ======================================================
// Actual function definition
// ======================================================


// ------------------------------------------------------
// Private Helpler
// ------------------------------------------------------
inline void CMatrix::swap(double &x, double &y)
{
	double temp = x;
	x = y;
	y = temp;
}

// ------------------------------------------------------
// constructors and destructors
// ------------------------------------------------------

/**
 * Copy Constructor\n
 * Creates a matrix with the same properties and values as another matrix.
 * Argument Description 
 * @param &refmatrix reference to another matrix 
 */
inline CMatrix::CMatrix (const CMatrix &refmatrix)
{
	_rows = refmatrix._rows;
	_cols = refmatrix._cols;

	_m = new double[_rows*_cols];
	memcpy(_m, refmatrix._m, _rows*_cols*sizeof(double));
}

/**
 * Constructor\n
 * If storage is zero, create a dynamically sized double-precision 
 * matrix with the number of rows and cols specified. If storage is 
 * not zero, use the memory it points to.
 * @param rows number of rows
 * @param cols number of cols
 * @param *storage
 */
inline CMatrix::CMatrix(int rows, int cols, double *storage) 
{
    _rows = rows;
	_cols = cols;

	if(0 == storage)
	{
	    _m = new double[rows*cols] ;
        memset( _m, 0, rows*cols * sizeof(double) ) ; 
	}
	else
	{
		_m = storage;
	}
}

/**
 * If storage is not zero, reorganizes the memory storage points 
 * to to have the number of rows and cols specified. If storage 
 * is zero, the appropriate amount of memory is allocated.
 * 
 * @param rows number of rows
 * @param cols number of columns
 * @param *storage storage pointer to memory for matrix
 */
inline void CMatrix::reSize (int rows, int cols, double *storage) 
{
	if(_m != NULL)
		delete[] _m ; 

    _rows = rows;
	_cols = cols;

	if(0 == storage)
	{
	    _m = new double[rows*cols] ;
        memset( _m, 0, rows*cols * sizeof(double) ) ; 
	}
	else
	{
		_m = storage;
	}	
}

/**
 * Destructor\n
 * clear the memory
 * @return
 */
inline CMatrix::~CMatrix (void) 
{
	if(0 != _m)
		delete[] _m;
}

// ------------------------------------------------------
// Accessors and Mutators
// ------------------------------------------------------
/**
 * Returns the number of rows in the matrix.
 * @return int: number of rows 
 */
inline int CMatrix::getRows (void) const
{
	return _rows;
}

/**
 * Returns the number of columns in the matrix.
 * @return int: number of columns 
 */
inline int CMatrix::getCols (void) const
{
	return _cols;
}


// ------------------------------------------------------
// Predicators
// ------------------------------------------------------
/**
 * Check whether the current matrix is valid
 * @return bool: true if the current matrix is valid and false otherwise
 */
inline bool CMatrix::isValid() const
{
	return (_rows > 0 || _cols >0 && 0 != _m);
}

/**
 * Check whether the current matrix is a square matrix
 * A square matrix has the same number of rows and columns
 *
 * @return bool: true if the current matrix is a square matrix and false otherwise
 */
inline bool CMatrix::isSquare() const
{
	return (_rows == _cols);
}


/**
 * Returns true if this matrix is symmetric, false otherwise.
 * A symmetric matrix has the same number of rows and columns (square). 
 * The value of element (i, j) must be the same as element (j, i) for all i and j.
 *
 * @return bool: true if the matrix is symmetric
 */
inline bool CMatrix::isSymmetric () const 
{
	if(!isSquare())
		return false;

	for(int i = 0; i< _rows; i++)
		for(int j=0; j< _cols; j++)
			if(i < j)
				if(EL(_m,i,j) != EL(_m,j,i))
					return false;
	return true;
}

/**
 * Returns true if this matrix is the identity matrix, false otherwise.
 * An identity matrix is symmtric matrix, and it has 1's on its diagonal and 0's everywhere else.
 *
 * @return bool: true if this matrix is the identity matrix, false otherwise.
 */
inline bool CMatrix::isIdentity () const 
{
	if(!isSquare())
		return false;

	for(int i = 0; i< _rows; i++)
		for(int j=0; j< _cols; j++)
			if(i == j)
				if(EL(_m,i,i) != 1)
					return false;
			else
				if(EL(_m,i,j) != 0)
					return false;
	return true;
}

/**
 * Check whether the matrix is compatible with the reference matrix
 * by "compatible", we mean both matrices are of the same dimention
 * same number of rows and columns, respectively.
 *
 * @param refmatrix the reference matrix
 * @return bool: true if the matrix is compatible with the refmatrix 
 */
inline bool CMatrix::isCompatible(const CMatrix& refmatrix) const
{
	return (isValid() && refmatrix.isValid() 
		&& _rows == refmatrix._rows && _cols == refmatrix._cols);
}


/**
 * Check whether the matrix can be right or left multiplied by the reference
 * matrix. (i.e.  m.cols = refvector.length)
 *
 * @param refvector reference to a vector
 * @return bool: true if the matrix can be right or left multiplied by the refvector
 */
inline bool CMatrix::isMultiplicable(const CVector& refvector) const
{
	return (isValid() && refvector.isValid()
		&& (_rows == refvector.length() || _cols == refvector.length()));
}

/**
 * Check whether the matrix can be right multiplied by the reference
 * matrix. (i.e.  m.cols = refmatrix.rows)
 *
 * @param refmatrix the reference matrix
 * @return bool: true if the matrix can be right multiplied by the refmatrix 
 */
inline bool CMatrix::isMultiplicable(const CMatrix& refmatrix) const
{
	return (isValid() && refmatrix.isValid()
		&& _cols == refmatrix._rows);
}

/**
 * Make the current matrix an identity matrix
 * @return reference to this
 */
inline CMatrix& CMatrix::identity(void)
{
	assert(isSquare());

	for(int i=0; i< _rows; i++)
		for(int j=0; j< _cols; j++)
			if(i == j)
				EL(_m, i, i) = 1;
			else
				EL(_m, i, j) = 0;
	return(*this);
}

/**
 * Returns a matrix that is the inverse of the current matrix.
 * @see invert
 * @return a new matrix that is the inverse of the current matrix
 */
inline CMatrix CMatrix::inverted (void) const 
{
	CMatrix mat(*this);
	return CMatrix(mat.invert());
} 

/**
 * Returns a matrix that is the transposition (switching of rows 
 * and columns) of the current matrix.
 * @see transpose
 * @return a new matrix that is the transpose of the current matrix
 */
inline CMatrix CMatrix::transposed (void) const 
{
	CMatrix mat(*this);
	return CMatrix(mat.transpose());
}


// ------------------------------------------------------
// Overloaded Operators 
// ------------------------------------------------------

/**
 * operator []: row access
 * @param i the index of the row starting from 0 
 * @return return the address pointing to the start of the row.
 * then m[i][j] will find the specific j element in that row
 */
inline double* CMatrix::operator[] (int i) const 
{
	assert(isValid() && (i >= 0 && i< _rows));
	// return the address pointing to the start of the row.
	// then m[i][j] will find the specific j element in that row
	return (&(_m[i*_cols]));

}

/**
 * operator =: assignment operator
 * Assigns value to all elements of this matrix.
 * @param value the value to assign
 * @return a reference to this
 */
inline CMatrix& CMatrix::operator= (double value) 
{
	assert(isValid());

	for(int i = 0; i<_rows*_cols; i++)
		_m[i] = value;
	return (*this);
}

/**
 * operator =: assignment operator
 * Assigns the elements in another matrix to this matrix. 
 * If the number of columns or the number of rows in the 
 * matrixes are not the same, an assertion error occurs.
 * @param &refmatrix reference to matrix whose elements are to be assigned to elements in this matrix
 * @return 
 */
inline CMatrix& CMatrix::operator= (const CMatrix &refmatrix) 
{
	assert(isCompatible(refmatrix));

	memcpy(_m, refmatrix._m, _rows*_cols*sizeof(double));
	return (*this);
}

/**
 * operator - [unary]
 * @return a matrix in which the elements of the current matrix have been multiplied by -1.
 */
inline CMatrix CMatrix::operator- (void) const 
{
	assert(isValid());
	CMatrix mat(*this);
	for(int i = 0; i<_rows*_cols; i++)
		mat._m[i] = -mat._m[i];
	return CMatrix(mat);
}


/**
 * operator +
 * @param refmatrix the reference matrix to add
 * @return a matrix that is the sum of the current matrix and refmatrix.
 */
inline CMatrix CMatrix::operator+ (const CMatrix& refmatrix) const 
{
	assert(isCompatible(refmatrix));
	CMatrix mat(*this);

	for(int i = 0; i<_rows*_cols; i++)
		mat._m[i] += refmatrix._m[i];

	return CMatrix(mat);	
}

/**
 * operator -
 * @param refmatrix the reference matrix to subtract from the current matrix
 * @return a matrix that is the difference of the current matrix and refmatrix.
 */
inline CMatrix CMatrix::operator- (const CMatrix& refmatrix) const 
{
	assert(isCompatible(refmatrix));
	CMatrix mat(*this);

	for(int i = 0; i<_rows*_cols; i++)
		mat._m[i] -= refmatrix._m[i];

	return CMatrix(mat);		
}

/**
 * operator *
 * @param dbl double-precision multiplier or divider
 * @return a matrix that is the product of the current matrix and dbl.
 */
inline CMatrix CMatrix::operator* (double dbl) const 
{
	assert(isValid());
	CMatrix mat(*this);

	for(int i = 0; i<_rows*_cols; i++)
		mat._m[i] *= dbl;

	return CMatrix(mat);	
}

/**
 * operator /
 * @param dbl double-precision multiplier or divider
 * @return a matrix that is the quotient of the current matrix and dbl.
 */
inline CMatrix CMatrix::operator/ (double dbl) const 
{
	assert(isValid());
	CMatrix mat(*this);

	for(int i = 0; i < _rows*_cols; i++)
		mat._m[i] *= dbl;

	return CMatrix(mat);	
}

/**
 * operator +=
 * Adds the elements of refmatrix to the corresponding elements of the current matrix.
 * @param refmatrix reference to a matrix
 * @return a reference to this
 */
inline CMatrix& CMatrix::operator+= (const CMatrix& refmatrix) 
{
	assert(isCompatible(refmatrix));
	for(int i=0; i < _rows*_cols; i++)
		_m[i] += refmatrix._m[i];
	return (*this);	
}

/**
 * operator -=
 * Subtracts the elements of refmatrix from the corresponding 
 * @param refmatrix reference to a matrix
 * @return a reference to this
 */
inline CMatrix& CMatrix::operator-= (const CMatrix& refmatrix) 
{
	assert(isCompatible(refmatrix));
	for(int i=0; i < _rows*_cols; i++)
		_m[i] += refmatrix._m[i];
	return (*this);	
}

/**
 * operator *=
 * Multiplies the elements of the current matrix by dbl
 * @param dbl a double valued multiplier 
 * @return reference to this
 */
inline CMatrix& CMatrix::operator*= (double dbl) 
{
	assert(isValid());
	for(int i=0; i < _rows*_cols; i++)
		_m[i] *= dbl;
	return (*this);	
}

/**
 * operator /= 
 * Divides the elements of the current matrix by dbl.
 * @param dbl a double valued divider
 * @return reference to this
 */
inline CMatrix& CMatrix::operator/= (double dbl) 
{
	assert(isValid());
	for(int i=0; i < _rows*_cols; i++)
		_m[i] /= dbl;
	return (*this);		
}

/**
 * operator ==
 * @param refmatrix reference to a matrix
 * @return true if all elements of refmatrix are equal to the
 * corresponding elements of the current matrix, or false otherwise
 */
inline bool CMatrix::operator== (const CMatrix& refmatrix) const 
{
	assert(isValid() && refmatrix.isValid());

	if(	_rows != refmatrix._rows || _cols != refmatrix._cols)
		return false;

	for(int i = 0; i< _rows*_cols; i++)
		if(_m[i] != refmatrix._m[i])
			return false;
	
	return true;
}

/**
 * operator !=
 * @param refmatrix reference to a matrix
 * @return true if any element of refmatrix is not equal to 
 * the corresponding element of the current matrix, or false otherwise
 */
inline bool CMatrix::operator!= (const CMatrix& refmatrix) const
{
	return !(*this == refmatrix);
}

/**
 * operator <
 * This operator is defined so this class can be used in STL containers.
 * @param refmatrix reference to a matrix
 * @return true if any element of refmatrix is less than the corresponding
 * element of the current matrix. not really meaningful here.
 */
inline bool CMatrix::operator< (const CMatrix& refmatrix) const 
{	
	assert(isCompatible(refmatrix));
	for(int i = 0; i< _rows*_cols; i++)
		if(_m[i] >= refmatrix._m[i])
			return false;
	return true;
}


#endif
//@}