/** =================================================================\n
 *
 * MathLib\n
 * @file Quaternion.h
 * @brief Quaternion System structure definition file
 * 
 * Primary Author: Howard Zhou (howardz@cc.gatech.edu)\n
 *
 * (C) Copyright Howard Zhou 2003\n
 * (C) Copyright Georgia Institute of Technology, 2003\n
 *
 * ==================================================================\n
 * Quaternion Lib
 *
 * @version 1.1\n 
 *     Major organization change, moved most of function defintion to\n
 *     the .cpp file to make class declaration more clear.
 * @date 2003-05-15
 * @author howardz
 * 
 * @version 1.0\n
 *     Initial Revision (in C++)
 * @date 2003-04-15
 * @author howardz
 *
 * @version 0.0\n
 *     Initial Version (in C)
 * @date 1997-09-01  
 * @author Jeff Lander   
 *
 * @todo none
 * @bug none 
 * @warning none
 *
 * ===================================================================\n
 */

#include "Quaternion.h"

#include <iostream>
#include <iomanip>
#include <math.h>

/// @namespace using standard namespace
using namespace std;

// ======================================================
// Actual function definition
// ======================================================

// ------------------------------------------------------
// constructors and destruct
// ------------------------------------------------------	

/**
 * Constructor\n
 * Converts rotational matrix to quaternion 
 * @param &m reference to a rotational matrix 
 */
CQuaternion::CQuaternion(const CRotationMatrix& m)
{
	double ww, xx, yy;
	// tolerance of error (for == 0 case)
	double e = 0.0001, temp;

	ww = 0.25*(1+m[0][0] + m[1][1] + m[2][2]);

	if(ww > e)
	{
		_w = sqrt(ww);
		temp = 1/_w * 0.25;
		_x = (m[1][2]-m[2][1])*temp;
		_y = (m[2][0]-m[0][2])*temp;
		_z = (m[0][1]-m[1][0])*temp;
	}
	else
	{
		_w = 0;
		xx = -0.5*(m[1][1] + m[2][2]);
		if(xx > e)
		{
			_x = sqrt(xx);
			temp = 1/_x*0.5;
			_y = m[0][1]*temp;
			_z = m[0][2]*temp;
		}
		else
		{	_x = 0;
			yy = -0.5 *(1-m[2][2]);
			if(yy > e)
			{
				_y = sqrt(yy);
				_z = m[2][2]*0.5/_y;
			}
			else
			{
				_y = 0;
				_z = 1;
			}
		}
	}
	_x = -_x, _y = -_y, _z = -_z;
}

/**
 * Constructor \n
 * Gets the shortest rotation that aligns vector a with vector b
 * @param &a reference to a vector 
 * @param &b reference to a vector
 */
CQuaternion::CQuaternion(const CVector3& a, const CVector3& b)
{
	// CVector3 a1 = a.normalized();
	// CVector3 b1 = b.normalized();
	CVector3 axis = a.cross(b);					// get the axis of the rotation
	double angle = acos((a*b)/(a.magnitude()*b.magnitude()));
												// get the angle
	CQuaternion q(angle, axis);
	_w = q._w, _x = q._x, _y = q._y, _z = q._z; 
}


// ------------------------------------------------------
// Accessors and Mutators
// ------------------------------------------------------

/**
 * Gets the (non-unique) rotational matrix corresponding to this quaternion\n
 * M = \n
 * | 1-2yy-2zz    2xy-2wz    2xz+2wy |\n
 * |   2xy+2wz  1-2xx-2zz    2yz-2wx |\n
 * |   2xz-2wy    2yz+2wx  1-2xx-2yy |\n
 * assuming the quaternion is normalized
 * @return a rotation matrix corresponding to this quaternion
 * notice that this rotational matrix is not unique
 */
CRotationMatrix CQuaternion::getRotationMatrix(void) const
{
	double x2, y2, z2, xx2, yy2, zz2, wx2, wy2, wz2, xy2, xz2, yz2;
	double *d = new double[9];
	x2 = _x+_x;
	y2 = _y+_y;
	z2 = _z+_z;
	xx2 = x2*_x;
	yy2 = y2*_y;
	zz2 = z2*_z;
	wx2 = x2*_w;
	wy2 = y2*_w;
	wz2 = z2*_w;
	xy2 = x2*_y;
	xz2 = x2*_z;
	yz2 = y2*_z;


	d[0] = 1 - yy2 - zz2;
	d[1] = xy2 - wz2;
	d[2] = xz2 + wy2;
	d[3] = xy2 + wz2;
	d[4] = 1 - xx2 - zz2;
	d[5] = yz2 - wx2;
	d[6] = xz2 - wy2;
	d[7] = yz2 + wx2;
	d[8] = 1 - xx2 - yy2;
	return CRotationMatrix(d);
}

	

/**
 * Gets the euler angles about three principle axis (determined by rotationType argument)\n
 * where M is the rotational matrix\n
 * phi   = atan(M[2][3], M[3][3])\n
 *       = atan(2(q2q3 + q0q1), (q0q0 - q1q1 - q2q2 + q3q3))\n
 * theta = asin(-M[1][3])\n
 *       = asin(-2(q1q3-q0q2))\n
 * psi   = atan(M[1][2], M[1][1])\n
 *       = atan(2(q1q2+q0q3), (q0q0+q1q1-q2q2-q3q3))\n
 * notice because of the ambiguity introduced by\n
 * cos(t) = +- sqrt(1-sin(t)*sin(t))
 * the solution won't be unique
 * @param rotationType the rotation type, XYZ or ZXY
 * @return the vector contains 3 Euler angles about the three principle axis
 * @todo change CVector3 into CEulerAngle class
 */
CVector3 CQuaternion::getEulerAngles(int rotationType) const
{
	double q0q0, q1q1, q2q2, q3q3, q0q1, q0q2, q0q3, q1q2, q1q3, q2q3;
	double rx, ry, rz;
	// double sx, sy, sz, cx, cy, cz;
	// double temp, temp1;

	// return (this->getRotationMatrix()).getEulerAngles(ZXY);
	switch(rotationType)
	{
	case CHANNEL_TYPE_RXYZ:
		q0q0 = _w*_w;
		q1q1 = _x*_x;
		q2q2 = _y*_y;
		q3q3 = _z*_z;
		q0q1 = _w*_x;
		q0q2 = _w*_y;
		q0q3 = _w*_z;
		q1q2 = _x*_y;
		q1q3 = _x*_z;
		q2q3 = _y*_z;
		ry   = asin(2*(q1q3+q0q2));
		
		// If cos(ry) is zero, then we must avoid dividing by zero. It 
		// also becomes impossible to distinguish roll from yaw. 
		// adopting the convention that the yaw angle is 0
		// temp = fabs(ry-PI*0.5);
		// temp1 = EPISOLON;
		//if(EPISOLON > fabs(ry-PI*0.5))
		//{
		//	rz = 0;
		//	rx = asin(2*(q2q3 + q0q1));
		//}
		//else
		//{
			rx   = atan2(-(2*(q2q3-q0q1)),(q0q0-q1q1-q2q2+q3q3));
			rz   = atan2(-(2*(q1q2-q0q3)),(q0q0+q1q1-q2q2-q3q3));
		//}
		break;
	case CHANNEL_TYPE_RZXY:
		q0q0 = _w*_w;
		q1q1 = _x*_x;
		q2q2 = _y*_y;
		q3q3 = _z*_z;
		q0q1 = _w*_x;
		q0q2 = _w*_y;
		q0q3 = _w*_z;
		q1q2 = _x*_y;
		q1q3 = _x*_z;
		q2q3 = _y*_z;
		rx   = asin(2*(q2q3+q0q1));
		//if(EPISOLON > fabs(rx-PI*0.5))
		//{
		//	rz = 0;
		//	ry = asin(2*(q1q2+q0q3));
		//}
		//else
		//{
			ry   = atan2(-(2*(q1q3-q0q2)),(q0q0-q1q1-q2q2+q3q3));
			rz   = atan2(-(2*(q1q2-q0q3)),(q0q0-q1q1+q2q2-q3q3));
		// }
	default:
		break;
	};

	return CVector3(rx, ry, rz);
}


// 

/**
 * Converts a set of Euler angles to a Quaternion\n
 * Takes in a set of 3 Euler angles (in radian) representing the 3 rotation performed\n
 * around x, y, z axis according to the order defined by CHANNEL_TYPE_\n
 *
 * This is a more optimized version, but less clear\n
 * @see eulerToQuaternion2: a conceptually more clear version\n
 *
 * @param rx angle in radian about x axis
 * @param ry angle in radian about y axis
 * @param rz angle in radian about z axis
 * @param rotationType rotation order: XYZ or ZXY
 * @return a set of Euler angles corresponding to this Quaternion
 */
CQuaternion& CQuaternion::eulerToQuaternion( double rx, double ry, double rz, int rotationType)
{
	double tx, ty, tz, cx, cy, cz, sx, sy, sz, cc, cs, sc, ss;
	// get the half angles
	tx = rx*0.5;
	ty = ry*0.5;
	tz = rz*0.5;
	cx = cos(tx);
	cy = cos(ty);
	cz = cos(tz);
	sx = sin(tx);
	sy = sin(ty);
	sz = sin(tz);

	if(rotationType == CHANNEL_TYPE_RXYZ || rotationType == CHANNEL_TYPE_SRT)
	{
		cc = cy * cz;
		cs = cy * sz;
		sc = sy * cz;
		ss = sy * sz;

		/*
		_w = cx*cc + sx*ss;		// cx*cy*cz + sx*sy*sz
		_x = sx*cc - cx*ss;		// sx*cy*cz - cx*sy*sz
		_y = cx*sc + sx*cs;		// cx*sy*cz + sx*cy*sz
		_z = cx*cs - sx*sc;		// cx*cy*sz - sx*sy*cz
		*/
		_w = cx*cc - sx*ss;		// cx*cy*cz - sx*sy*sz
		_x = sx*cc + cx*ss;		// sx*cy*cz + cx*sy*sz
		_y = cx*sc - sx*cs;		// cx*sy*cz - sx*cy*sz
		_z = cx*cs + sx*sc;		// cx*cy*sz + sx*sy*cz
	}
	else if(rotationType == CHANNEL_TYPE_RZXY || rotationType == CHANNEL_TYPE_TRANS)
	{
		cc = cy * cz;
		cs = cy * sz;
		sc = sy * cz;
		ss = sy * sz;
	
		/*
		_w = cx*cc + sx*ss;		// cx*cy*cz + sx*sy*sz
		_x = cx*ss + sx*cc;		// cx*sy*sz + sx*cy*cz
		_y = cx*sc - sx*cs;		// cx*sy*cz - sx*cy*sz
		_z = cx*cs - sx*sc;		// cx*cy*sz - sx*sy*cz
		*/

		_x = sx*cc - cx*ss;		// sx*cy*cz - cx*sy*sz
		_y = cx*sc + sx*cs;		// cx*sy*cz + sx*cy*cz
		_z = sx*sc + cx*cs;		// sx*sy*cz + cx*cy*sz
		_w = cx*cc - sx*ss;		// cx*cy*cz - sx*sy*sz
	}
	// insure the quaternion is normalized
	// not necessary, because quaternion multiplication preserver magnitude
	return (*this);
}

/**
 * Converts a set of Euler angles to a Quaternion\n
 * Takes in a set of 3 Euler angles (in radian) representing the 3 rotation performed\n
 * around x, y, z axis according to the order defined by CHANNEL_TYPE_\n
 *
 * This is a more clear version, but less optimized\n
 * @see eulerToQuaternion: a heavily optimized version\n
 *
 * @param rx angle in radian about x axis
 * @param ry angle in radian about y axis
 * @param rz angle in radian about z axis
 * @param rotationType rotation order: XYZ or ZXY
 * @return a set of Euler angles corresponding to this Quaternion
 */
CQuaternion& CQuaternion::eulerToQuaternion2( double rx, double ry, double rz, int rotationType)
{
	// get the half angles
	double ti = rx*0.5;
	double tj = ry*0.5;
	double tk = rz*0.5;

	CQuaternion qx = CQuaternion(cos(ti), sin(ti), 0, 0.00);
	CQuaternion qy = CQuaternion(cos(tj), 0, sin(tj), 0.00);
	CQuaternion qz = CQuaternion(cos(tk), 0, 0, sin(tk));
	CQuaternion qf;

	// remember
	if(rotationType == CHANNEL_TYPE_RXYZ)
		qf = qz*qy*qx;
	else if(rotationType == CHANNEL_TYPE_RZXY)
		qf = qy*qx*qz;

	// insure the quaternion is normalized
	// not necessary, because quaternion multiplication preserver magnitude
	_w = qf._w;
	_x = qf._x;
	_y = qf._y;
	_z = qf._z;
	return (*this);
}

// ------------------------------------------------------
// Output or Test
// ------------------------------------------------------

/**
 * Stand along member display function
 */
void CQuaternion::display() const
{	
	cout << " [";
	cout << setprecision(3) << setw(6) << setiosflags (ios_base::fixed | ios_base::right)<< _w << " ";
	cout << setprecision(3) << setw(6) << setiosflags (ios_base::fixed | ios_base::right)<< _x << " ";
	cout << setprecision(3) << setw(6) << setiosflags (ios_base::fixed | ios_base::right)<< _y << " ";
	cout << setprecision(3) << setw(6) << setiosflags (ios_base::fixed | ios_base::right)<< _z << " ";
	cout << " ] ";
}


/**
 * General test function for CMatrix class
 * put anything you want to test here.
 */
void CQuaternion::test(void) const
{
	// quaternion basic constructor and display function
	char* hr = "\nCQuaternion class Test \n=============================================================\n";
	CQuaternion q1, q2;
	cout << "\nFirst, check for constructors and display functions" << hr;
	cout << q1 << endl;
	q2.display();
	cout << endl;

	// quaternion multiplication and normalization
	cout << "\n\nNow let's try normalized quaternion multiplication" << hr;
	CQuaternion q3(2, 2, 2, 1);
	CQuaternion q4(3, 3, -1, 0);
	CQuaternion q5 = q3*q4;
	cout << "\nUn-normalized multiplication : " << q3 << " * " << q4 << " = " << q5;
	cout << "\nNormalize the product : " << q5.normalize();
	q3.normalize();
	q4.normalize();
	q5 = q3*q4;
	cout << "\nNormalized multiplication : " << q3 << " * " << q4 << " = " << q5;
	cout << "\nNormalize the product : " << q5.normalize();

	// Quaternion multiplication
	cout << "\n\nTest of a special case to check whether the multiplication is correct" << hr;
	CQuaternion q6(0, 0, 1, 0);
	CQuaternion q7(0, 0, 0, 1);
	CQuaternion q8 = q6*q7;
	cout << "\n" << q6 << " * " << q7 << " = " << q8;
	cout << "\n" << q6 << " * " << q8 << " = " << q6*q8;
	cout << "\n" << q8 << " * " << q6 << " = " << q8*q6;
	cout << "\n" << q7 << " * " << q8 << " = " << q7*q8;
	cout << "\n" << q8 << " * " << q7 << " = " << q8*q7;


	// Quaternion to rotation
	cout << "\nNow some test about angles and rotational matrix" << hr;	
	CQuaternion q9 = CQuaternion(RAD(180), CVector3(0, 1, 0));
	cout << "\nQuaternion corresponding to 180 degree around y axis (180|0,1,0): " << q9;
	cout << "\nNow get back the axis : " << q9.getAxis();
	cout << "\nand the angle : " << DEG(q9.getAngle());
	cout << "\nand the rotation matrix M : \n" << q9.getRotationMatrix() << endl;

	CVector3 v = CVector3(12, 34, 2);
	cout << "\n\nNow let's see whether the quaternion multiplication gives us the same result as multipling by the matrix";
	cout << "\nv' = q.inverted() * v * q : " << (q9.inverted() * CQuaternion(v) * q9).getVector() << endl;
	cout << "\nv' = M * v : " << q9.getRotationMatrix() * v << endl;


	// Euler Angle to Quaternion
	cout << "\n\nNow the useful part, Euler to Quaternion" << hr;
	cout << "\nlet rx = 20, ry = 40, rz = 75, what will be the result :";
	cout << "\nThis should be the rotationMatrix (XYZ): \n" <<  CRotationMatrix(RAD(20), RAD(40), RAD(75), XYZ);
	cout << "\nThis should be the rotationMatrix (ZXY): \n" <<  CRotationMatrix(RAD(20), RAD(40), RAD(75), ZXY);
	CQuaternion q10, q11;
	CRotationMatrix m10, m11;
	q10.eulerToQuaternion(RAD(20), RAD(40), RAD(75), CHANNEL_TYPE_RXYZ);
	cout << "\nNow it's quaternion's turn (XYZ):" << q10;
	cout << "\nNormalized ? magnitude : " << q10.magnitude();
	cout << "\nWhat is the underlying matrix M : \n" << q10.getRotationMatrix() << endl;
	q11.eulerToQuaternion(RAD(20), RAD(40), RAD(75), CHANNEL_TYPE_RZXY);
	cout << "\nNow it's quaternion's turn (ZXY):" << q11;
	cout << "\nNormalized ? magnitude : " << q11.magnitude();
	cout << "\nWhat is the underlying matrix M : \n" << q11.getRotationMatrix() << endl;
	q10.eulerToQuaternion2(RAD(20), RAD(40), RAD(75), CHANNEL_TYPE_RXYZ);
	cout << "\nNow it's quaternion's turn (XYZ):" << q10;
	cout << "\nNormalized ? magnitude : " << q10.magnitude();
	m10 = q10.getRotationMatrix();
	cout << "\nWhat is the underlying matrix M : \n" << m10 << endl;
	q11.eulerToQuaternion2(RAD(20), RAD(40), RAD(75), CHANNEL_TYPE_RZXY);
	cout << "\nNow it's quaternion's turn (ZXY):" << q11;
	cout << "\nNormalized ? magnitude : " << q11.magnitude();
	m11 = q11.getRotationMatrix();
	cout << "\nWhat is the underlying matrix M : \n" << m11 << endl;

	// now let's see if from the rotational matrix we can get back the quaternion
	cout << "\nNow what is q if we convert q10.RotationMatrix() to q :\n";
	CQuaternion q(m10);
	cout << q << endl;
	q = CQuaternion(m11);
	cout << "\nNow what is q if we convert q10.RotationMatrix() to q :\n";
	cout << q << endl;


	// Spherical Linear Interpolation Between two Quaternions
	cout << "\n\nSpherical Linear Interpolation Between two Quaternions" << hr;
	CQuaternion q12(0, 1, 0, 0);
	CQuaternion q13(0, 0, 1, 0);
	CQuaternion q14 = slerp(q12, q13, 0.5);
	cout << "\nSlerp for " << q12 << " and " << q13 << " is : " << q14 << endl;

	// now test the conversion from quaternion to rotation
	cout << "\n\nnow test the conversion from quaternion to Euler angles (indirectly from RotationMatrix)" << hr;
	cout << "\nUsing q10 as an example" << endl;
	cout << "\nThe result should be (20, 40, 75) = (" << RAD(20) << ", " << RAD(40) << ", " << RAD(75) << "): " << q10.getRotationMatrix().getEulerAngles(XYZ);
	cout << "\nUsing q11 as an example" << endl;
	cout << "\nThe result should be (20, 40, 75) = (" << RAD(20) << ", " << RAD(40) << ", " << RAD(75) << "): " << q11.getRotationMatrix().getEulerAngles(ZXY);
	cout << "\n\nnow test the conversion from quaternion to Euler angles (direct)" << hr;
	cout << "\nUsing q10 as an example" << endl;
	cout << "\nThe result should be (20, 40, 75) = (" << RAD(20) << ", " << RAD(40) << ", " << RAD(75) << "): " << q10.getEulerAngles(CHANNEL_TYPE_RXYZ);
	cout << "\nUsing q11 as an example" << endl;
	cout << "\nThe result should be (20, 40, 75) = (" << RAD(20) << ", " << RAD(40) << ", " << RAD(75) << "): " << q11.getEulerAngles(CHANNEL_TYPE_RZXY);

	CRotationMatrix rm(RAD(18), RAD(90), RAD(20), XYZ);
	CQuaternion q15;
	q15.eulerToQuaternion(RAD(18), RAD(90), RAD(20), CHANNEL_TYPE_RXYZ);
	cout << "\n\nnow test the conversion from quaternion to Euler angles (direct)" << hr;
	cout << "\n\nthe rotational Matrix \n" << rm;
	cout << "\n\nthe quaternion \n" << q15;
	cout << "\nThe result should be (18, 90, 20, XYZ) = (" << RAD(18) << ", " << RAD(90) << ", " << RAD(20) << "):\n " << q15.getEulerAngles(CHANNEL_TYPE_RXYZ);
	CQuaternion q16;
	q16.eulerToQuaternion(RAD(90), RAD(70), RAD(20), CHANNEL_TYPE_RZXY);
	cout << "\n\nnow test the conversion from quaternion to Euler angles (direct)" << hr;
	cout << "\n\nthe rotational Matrix \n" << q16.getRotationMatrix();
	cout << "\n\nthe quaternion \n" << q16;
	cout << "\nThe result should be (90, 70, 20, ZXY) = (" << RAD(90) << ", " << RAD(70) << ", " << RAD(20) << "):\n " << q16.getEulerAngles(CHANNEL_TYPE_RZXY);

	CQuaternion q17;
	q17.eulerToQuaternion(RAD(20), RAD(40), RAD(75), CHANNEL_TYPE_RXYZ);
	cout << "\nq17 (1) : " << q17;
	// q17.eulerToQuaternion2(RAD(20), RAD(40), RAD(75), CHANNEL_TYPE_RXYZ);	
	// cout << "\nq17 (2) : " << q17;
	// now test whether quaternion rotation will prouce the same result as Euler angles
	cout << "\n\nnow test whether quaternion rotation will prouce the same result as Euler angles"<< hr;
	cout << "\nCreate a random vector (67, 14, 20)" << endl;
	CVector3 v1(67, 14, 20);
	cout << "Let's rotate it using rx = 20, ry = 40, rz = 75 :" << endl;
	cout << "\nUsing Euler :\n" << CRotationMatrix(RAD(20), RAD(40), RAD(75), XYZ) * v1;
	cout << "\nUsing Quaternion :\n" << q17.rotate(v1);
	cout << "\nThe above should be the same." << endl;
}


// ------------------------------------------------------
// Friend functions
// ------------------------------------------------------
/**
 * operator << 
 * @param os reference to the output stream
 * @param q reference to a quaternion
 * @return reference to the output stream
 */
ostream& operator<< (ostream& os, CQuaternion& q)
{	
	os << " [";
	os << setprecision(3) << setw(6) << setiosflags (ios_base::fixed | ios_base::right)<< q._w << " ";
	os << setprecision(3) << setw(6) << setiosflags (ios_base::fixed | ios_base::right)<< q._x << " ";
	os << setprecision(3) << setw(6) << setiosflags (ios_base::fixed | ios_base::right)<< q._y << " ";
	os << setprecision(3) << setw(6) << setiosflags (ios_base::fixed | ios_base::right)<< q._z << " ";
	os << " ] ";
	return os;
}

/**
 * SLERP: Spherical Linear Interpolation Between two Quaternions\n
 * Adjust the DELTA constant to play with the LERP vs. SLERP level
 * @param q1 reference to a quaternion
 * @param q2 reference to a quaternion
 * @param blend blend factor
 * @return a quaternion resulting from slerp using the given blend factor
 */
CQuaternion slerp(CQuaternion& q1, CQuaternion& q2, float blend)
{
	double omega,cosom,sinom,scale0,scale1, x, y, z, w;

	// use the dot product to get the cosine of the angle between the
	// quaternions
	cosom = q1._x * q2._x + 
			q1._y * q2._y + 
			q1._z * q2._z + 
			q1._w * q2._w; 

	// check a couple of special cases. 
	// make sure the two quaternions are not exactly opposite? (within a little slop)
	if ((1.0 + cosom) > DELTA)
	{
		// are they more than a little bit different? avoid a divided by zero and lerp if not
		if ((1.0 - cosom) > DELTA) {
			// yes, do a slerp
			omega = acos(cosom);
			sinom = sin(omega);
			scale0 = sin((1.0 - blend) * omega) / sinom;
			scale1 = sin(blend * omega) / sinom;
		} else {
			// not a very big difference, do a lerp
			// when q1 and q2 are very close, the omega -> 0 and sin(omega) -> 0
			// sin((1 - blend)*omega) -> (1-blend)*omega, and sin(blend*omega)-> blend*omega
			// hence slerp(q1, q2: blend) = (1-blend)*q1 + blend*q2
			scale0 = 1.0 - blend;
			scale1 = blend;
		}
		x = scale0 * q1._x + scale1 * q2._x;
		y = scale0 * q1._y + scale1 * q2._y;
		z = scale0 * q1._z + scale1 * q2._z;
		w = scale0 * q1._w + scale1 * q2._w;
	} else {
		// the quaternions are nearly opposite so to avoid a divided by zero error
		// calculate a perpendicular quaternion and slerp that direction
		x = -q2._y;
		y = q2._x;
		z = -q2._w;
		w = q2._z;
		scale0 = sin((1.0 - blend) * (float)HALF_PI);
		scale1 = sin(blend * (float)HALF_PI);
		x = scale0 * q1._x + scale1 * x;
		y = scale0 * q1._y + scale1 * y;
		z = scale0 * q1._z + scale1 * z;
		w = scale0 * q1._w + scale1 * w;
	}
	return CQuaternion(w, x, y, z);
}
//@}