/** =================================================================\n
 *
 * MathLib\n
 * @file Quaternion.h
 * @brief Quaternion System structure definition file
 * 
 * Primary Author: Howard Zhou (howardz@cc.gatech.edu)\n
 *
 * (C) Copyright Howard Zhou 2003\n
 * (C) Copyright Georgia Institute of Technology, 2003\n
 *
 * ==================================================================\n
 * Quaternion Lib
 *
 * @version 1.1\n 
 *     Major organization change, moved most of function defintion to\n
 *     the .cpp file to make class declaration more clear.
 * @date 2003-05-15
 * @author howardz
 * 
 * @version 1.0\n
 *     Initial Revision (in C++)
 * @date 2003-04-15
 * @author howardz
 *
 * @version 0.0\n
 *     Initial Version (in C)
 * @date 1997-09-01  
 * @author Jeff Lander   
 *
 * @todo none
 * @bug none 
 * @warning none
 *
 * ===================================================================\n
 */

#ifndef __QUATERNION_H__
#define __QUATERNION_H__

#include <iostream>
#include <iomanip>

#include "Vector.h"
#include "RotationMatrix.h"

#ifndef EPISOLON
#define EPISOLON 0.0000001
#endif

#ifndef PI
#define PI 3.14159265358979323846
#endif

#ifndef HALF_PI
#define HALF_PI 1.57079632679489661923
#endif

#ifndef DELTA
#define DELTA 0.0001
#endif


#ifndef RADIAN_PER_DEGREE
#define RADIAN_PER_DEGREE 0.0174532925199432957
#define DEGREE_PER_RADIAN 57.295779513082320877
#define RAD(x) ((x)*RADIAN_PER_DEGREE)
#define DEG(x) ((x)*DEGREE_PER_RADIAN)
#endif

// ------ Channel Definitions -------------------------------------------------------
/// @def Channel Definitions
#ifndef CHANNEL_TYPE
#define CHANNEL_TYPE				0		// DUMMY
#define CHANNEL_TYPE_NONE			0		// NO CHANNEL APPLIED
#define CHANNEL_TYPE_SRT			1		// SCALE ROTATION AND TRANSLATION
#define CHANNEL_TYPE_TRANS			2		// CHANNEL HAS TRANSLATION (X Y Z) ORDER,
											//			   ROTATION (Z X Y) ORDER
#define CHANNEL_TYPE_RXYZ			4		// ROTATION (RX RY RZ) ORDER
#define CHANNEL_TYPE_RZXY			8		// ROTATION (RZ RX RY) ORDER
#define CHANNEL_TYPE_RYZX			16		// ROTATION (RY RZ RX) ORDER
#define CHANNEL_TYPE_RZYX			32		// ROTATION (RZ RY RX) ORDER
#define CHANNEL_TYPE_RXZY			64		// ROTATION (RX RZ RY) ORDER
#define CHANNEL_TYPE_RYXZ			128		// ROTATION (RY RX RZ) ORDER
#define CHANNEL_TYPE_S				256		// SCALE ONLY
#define CHANNEL_TYPE_T				512		// TRANSLATION ONLY (X Y Z) ORDER
#define CHANNEL_TYPE_INTERLEAVED	1024	// THIS DATA STREAM HAS MULTIPLE CHANNELS
#endif
// ----------------------------------------------------------------------------------

/// @namespace using standard namespace
using namespace std;

// ==================================================================================
/**
 * @class CQuaternion Quaternion.h Quaternion.h
 * @brief Quaternion class\n
 * see http://www.cc.gatech.edu/dvfx/readings/shoemake-s85.pdf for detailed info,
 * but notice that the above paper used a different convention as the following code
 * 
 * @ingroup Lib
 *
 * @par requirements
 *
 * @version 1.1\n 
 *     Major organization change, moved most of function defintion to\n
 *     the .cpp file to make class declaration more clear.
 * @date 2003-05-15
 * @author howardz
 * 
 * @version 1.0\n
 *     Initial Revision (in C++)
 * @date 2003-04-15
 * @author howardz
 *
 * @version 0.0\n
 *     Initial Version (in C)
 * @date 1997-09-01  
 * @author Jeff Lander   
 *
 *
 * @todo none
 * @bug none
 * @warning none
 */
// ================================================================================
class CQuaternion
{
protected:
	double _w;  ///< the scalar cos(a)
	double _x;  ///< the vector component
	double _y;  ///< the vector component
	double _z;  ///< the vector component

public:
	
	// ------------------------------------------------------
	// constructors and destructors
	// ------------------------------------------------------
	CQuaternion (void);
	CQuaternion (double w, double x, double y, double z);
	CQuaternion (const CQuaternion &q);
	CQuaternion (const CVector3& v);
	CQuaternion (double angle, const CVector3& axis);
	CQuaternion (const CVector3& a, const CVector3& b);
	CQuaternion (const CRotationMatrix& m);

	// default destructor	
	// ~CQuaternion(){}

	// ------------------------------------------------------
	// Predicates
	// ------------------------------------------------------
	bool isValid(void) const;

	// ------------------------------------------------------
	// Accessors and Mutators
	// ------------------------------------------------------
	void            set               (double w, double x, double y, double z);
    double          magnitude         (void) const; 
	CQuaternion&    normalize         (void);
	CQuaternion     normalized        (void) const;	
	CQuaternion&    identity          (void);
	CQuaternion&    invert            (void);
	CQuaternion     inverted          (void) const;	
	CVector3        rotate            (CVector3& v);
	double          getAngle          (void) const;
	CVector3        getAxis           (void) const;
	CVector3        getVector         (void);
	CRotationMatrix getRotationMatrix (void) const;
	CVector3        getEulerAngles    (int rotationType = CHANNEL_TYPE_RXYZ) const;
	CQuaternion&    eulerToQuaternion (double rx, double ry, double rz, int rotationType = CHANNEL_TYPE_RXYZ);
	CQuaternion&    eulerToQuaternion2(double rx, double ry, double rz, int rotationType = CHANNEL_TYPE_RXYZ);
	double          dot               (const CQuaternion& q);


	// ------------------------------------------------------
	// Overloaded Operators 
	// ------------------------------------------------------
	CQuaternion&    operator=  (const  CQuaternion& q);
	CQuaternion     operator*  (const  CQuaternion& q) const;
	CQuaternion     operator/  (const  CQuaternion& q) const;
	CQuaternion     operator-  (void)  const;
	CQuaternion     operator+  (const  CQuaternion& q) const;
	CQuaternion     operator-  (const  CQuaternion& q) const;
	CQuaternion     operator*  (double dbl) const;
	CQuaternion     operator/  (double dbl) const;
	CQuaternion&    operator+= (const  CQuaternion& q);
	CQuaternion&    operator-= (const  CQuaternion& q);
	CQuaternion&    operator*= (const  CQuaternion& q);
	CQuaternion&    operator/= (const  CQuaternion& q);
	CQuaternion&    operator*= (double dbl);
	CQuaternion&    operator/= (double dbl);
	bool            operator== (const  CQuaternion& q) const;
	bool            operator!= (const  CQuaternion& q) const;
	bool            operator<  (const  CQuaternion& q) const;

	// ------------------------------------------------------
	// Output or Test
	// ------------------------------------------------------
	void     display() const;
	void	    test() const;

	// ------------------------------------------------------
	// Friend functions
	// ------------------------------------------------------
    friend ostream& operator<< (ostream& os, CQuaternion& q);
	friend CQuaternion slerp(CQuaternion& q1,CQuaternion& q2,float blend);
};


// ======================================================
// Actual function definition
// ======================================================

// ------------------------------------------------------
// constructors and destructors
// ------------------------------------------------------
/**
 * Default Constructor\n
 * Makes it an identity quaternion
 */
inline CQuaternion::CQuaternion(void)
{
	_w = 1;   
	_x = 0;   
	_y = 0;   
	_z = 0;
}

/**
 * Copy constructor \n
 * @param &q reference to a quaternion
 */
inline CQuaternion::CQuaternion (const CQuaternion &q)
{
	_w = q._w;  
	_x = q._x;  
	_y = q._y;  
	_z = q._z;
}

/**
 * Constructor\n
 * @param w 
 * @param x 
 * @param y 
 * @param z 
 */
inline CQuaternion::CQuaternion(double w, double x, double y, double z)
{
	_w = w;  
	_x = x;  
	_y = y;   
	_z = z;
}

/**
 * Constructor (type casting) \n
 * Takes in an arbitray vector so that we can use quaternion multiplication
 * to rotate this vector
 * v' = q-1 v q;
 * @param &v reference to a 3D vector 
 */
inline CQuaternion::CQuaternion(const CVector3& v)
{
	_w = 0;
	_x = v[0];
	_y = v[1];
	_z = v[2];
}

/**
 * Constructor: angle-axis combination \n
 * Takes in an angle in radian and a vector representing the rotational axis
 * and convert it into a normalized quaternion
 * @param angle angle in radian 
 * @param axis the rotational axis
 */
inline CQuaternion::CQuaternion(double angle, const CVector3& axis)
{
	// don't forget normalize the axis
	CVector3 n = axis.normalized();
	double s = sin(angle*0.5);
	_w = cos(angle*0.5);
	_x = s*n[0];
	_y = s*n[1];
	_z = s*n[2];
}

// ------------------------------------------------------
// Predicators
// ------------------------------------------------------


/**
 * Checks whether the quaternion is valid
 * since w = cos(a/2) of some angle a, 
 * so we have -1 <= w = cos(a/2) <=1 
 * @return true if the quaternion is valid
 * @todo find a better way to check unnormalized input quaternions
 */
inline bool CQuaternion::isValid(void) const
{
	return true; // return (_w >= -1 && _w <= 1);
}


// ------------------------------------------------------
// Accessors and Mutators
// ------------------------------------------------------

/**
 * Sets all of its values
 * @param w 
 * @param x 
 * @param y 
 * @param z 
 */
inline void CQuaternion::set(double w, double x, double y, double z)
{
	_w = w;  
	_x = x;   
	_y = y;   
	_z = z;
}

/**
 * Magnitude
 * @return double: magnitude of the quaternion 
 */
inline double CQuaternion::magnitude(void) const 
{
	return sqrt(_w*_w + _x*_x + _y*_y+_z*_z);
}

/**
 * Normalize the quaternion\n
 * Before we do any rotation with the quaternion,
 * it has to be normalized
 * so that w = cos(a/2) and x, y, z = sin(a/2)*x',y',z'
 * where (x', y', z') is the axis of rotation
 * @return reference to the quaternion
 */
inline CQuaternion& CQuaternion::normalize(void)
{
	double mag = magnitude();
	_w /= mag; _x /= mag; _y /= mag; _z /= mag;
	return (*this);
}

/**
 * Returns a quaternion that is the normalized version of the current quaternion.
 * @return a quaternion that is the normalized version of the current quaternion
 */
inline CQuaternion CQuaternion::normalized (void) const 
{	
	double mag = magnitude();
	return CQuaternion(_w/mag, _x/mag, _y/mag, _z/mag);	
} 
 
/**
 * Make the current Quaternion an identity quaternion\n
 * identity q = <1, 0, 0, 0>\n
 * @return reference to the quaternion
 */
inline CQuaternion& CQuaternion::identity(void)
{
	_w = 1; _x = 0; _y = 0; _z = 0;
	return(*this);
}

/**
 * Inverts the current matrix.
 * @return reference to the quaternion
 */
inline CQuaternion& CQuaternion::invert (void)
{
	double mag2 = _w*_w + _x*_x + _y*_y + _z*_z;
	_w /= mag2, _x /= -mag2, _y /= -mag2, _z /= -mag2;
	return (*this);
}

/**
 * Returns a quaternion that is the inverse of the current quaternion.
 * @return a quaternion that is the inverse of the current quaternion
 */
inline CQuaternion CQuaternion::inverted (void) const 
{	
	double mag2 = _w*_w + _x*_x + _y*_y + _z*_z;
	return CQuaternion(_w/mag2, -_x/mag2, -_y/mag2, -_z/mag2);	
} 

/**
 * Rotates the vector and return the result in a vector
 * @param v reference to a vector
 * @return a vector resulting from rotating the original vector
 */
inline CVector3 CQuaternion::rotate(CVector3& v){
	// convert the vector into quaternion for later multiplication
	CQuaternion vq(v);
	// CQuaternion result = ((*this).inverted())*vq*(*this);
	CQuaternion result = (*this)*vq*((*this).inverted());
	return result.getVector();
}

/**
 * Converts a Quaternion to Axis Angle representation\n
 * Gets the rotational angle in radian\n
 * w = cos(a/2) -> a = 2*acos(w)
 * assuming the quaternion is normalized
 * As the order of rotations is important I am
 * using the Quantum Mechanics convention of (X,Y,Z)
 * a Yaw-Pitch-Roll (Y,X,Z) system would have to be adjusted
 * @return double: angle in radian
 */
inline double CQuaternion::getAngle(void) const
{
	double mag = magnitude();
	return acos(_w/mag)*2;
}

/**
 * Convert a Quaternion to Axis Angle representation\n
 * Gets the rotational axis
 * x, y, z = sin(a/2)* x', y', z'
 * assuming the quaternion is normalized
 * @return 
 */
inline CVector3 CQuaternion::getAxis(void) const{
	double mag = magnitude();
	double a = acos(_w/mag)*2;
	double scale = sin(a*0.5);
	return CVector3(_x/scale, _y/scale, _z/scale);
}

/**
 * Gets the vector after rotation is done
 * @see CQuaternion(CVector3 v) constructor
 * @return 
 */
inline CVector3 CQuaternion::getVector(void)
{
	return CVector3(_x, _y, _z);
}

/**
 * Inner product in Euclidean space, used to find the
 * distance between two quaternions
 * @param q const reference to a quaternion
 */
inline double CQuaternion::dot (const CQuaternion& q)
{
	return (_w*q._w + _x*q._x + _y*q._y + _z*q._z);
}
	
// ------------------------------------------------------
// Overloaded Operators 
// ------------------------------------------------------


/**
 * Assignment \n
 * operator = \n
 * Copies the values of another vector to the 
 * corresponding elements of this vector.
 * @param q reference to a quaternion
 * @return reference to this quaternion
 */
inline CQuaternion& CQuaternion::operator= (const CQuaternion& q) {
	assert(q.isValid());	
	_w = q._w; _x = q._x; _y = q._y; _z = q._z;
	return (*this);
}	

/**
 * operator * \n
 * Returns the product of the current quaternion and the ref quaternion.
 * q1*q2 = (w1,v1)*(w2,v2) = (w1*w2-v1*v2, w1*v2 + w2*v1 + v1 x v2)
 * @param &q reference to a quaternion
 * @return the product of the current quaternion and the ref quaternion
 */
inline CQuaternion CQuaternion::operator* (const CQuaternion &q) const
{
	assert(isValid() && q.isValid());
	double w = _w*q._w - _x*q._x - _y*q._y - _z*q._z;
	double x = _w*q._x + _x*q._w + _y*q._z - _z*q._y;
	double y = _w*q._y - _x*q._z + _y*q._w + _z*q._x;
	double z = _w*q._z + _x*q._y - _y*q._x + _z*q._w;
	return CQuaternion(w, x, y, z);
}


/**
 * operator / \n	
 * Returns the product of the current quaternion q0 and the ref quaternion q.
 * actually q1 = q0/q ==> q1 = q0*q.inverted() 
 * @param &q reference to a quaternion
 * @return the product of the current quaternion q0 and the ref quaternion q
 */
inline CQuaternion CQuaternion::operator/ (const CQuaternion &q) const
{
	return CQuaternion((*this)*q.inverted());
}

/**
 * operator - [unary] \n
 * @return a quaternion in which the elements of 
 * the current quaternion have been multiplied by -1.s
 */
inline CQuaternion CQuaternion::operator- (void) const {
	assert(isValid());
	return CQuaternion(-_w, -_x, -_y, -_z);
}

/**
 * operator + \n
 * I don't know why someone wants to add quaternions, so I will leave this operator here
 * in the convention of a complex number
 * @param q reference to a quaternion
 * @return a quaternion that is the sum of the current quaternion and ref quaternion
 */
inline CQuaternion CQuaternion::operator+ (const CQuaternion& q) const {
	assert(q.isValid());		
	return CQuaternion(_w + q._w, _x + q._x, _y + q._y, _z + q._z);
}


/**
 * operator - \n
 * @param q reference to a quaternion
 * @return a quaternion that is the difference of the current quaternion and ref quaternion
 */
inline CQuaternion CQuaternion::operator- (const CQuaternion& q) const {
	assert(q.isValid());
	return CQuaternion(_w + q._w, _x + q._x, _y + q._y, _z + q._z);
}

/**
 * operator * \n
 * @param dbl double as a multiplier
 * @return a quaternion that is the current quaternion multiplied the dbl value
 */
inline CQuaternion CQuaternion::operator* (double dbl) const{
	assert(isValid());
	return CQuaternion(_w*dbl, _x*dbl, _y*dbl, _z*dbl);
	
}

/**
 * operator / \n
 * @param dbl double as a divider
 * @return a quaternion that is the current quaternion divided the dbl value
 */
inline CQuaternion CQuaternion::operator/ (double dbl) const {
	assert(isValid());
	return CQuaternion(_w/dbl, _x/dbl, _y/dbl, _z/dbl);
}

/**
 * operator += \n
 * Adds the elements of refquaternion to the corresponding elements of the current quaternion
 * @param q reference to a quaternion
 * @return reference to this quaternion
 */
inline CQuaternion& CQuaternion::operator+= (const CQuaternion& q) {
	_w+=q._w; _x+=q._x; _y+=q._y; _z+=q._z;
	return (*this);
}


/**
 * operator -= \n
 * Subtracts the elements of refquaternion from the corresponding elements of the current quaternion.
 * @param q reference to a quaternion
 * @return reference to this quaternion
 */
inline CQuaternion& CQuaternion::operator-= (const CQuaternion& q) {
	_w-=q._w; _x-=q._x; _y-=q._y; _z-=q._z;
	return (*this);
}

/**
 * operator * \n	
 * Returns the product of the current quaternion and the ref quaternion.
 * q1*q2 = (w1,v1)*(w2,v2) = (w1*w2-v1*v2, w1*v2 + w2*v1 + v1 x v2)
 * @param &q reference to a quaternion
 * @return reference to this quaternion
 */
inline CQuaternion& CQuaternion::operator*= (const CQuaternion &q){
	assert(isValid() && q.isValid());
	CQuaternion temp;

	temp._w = _w*q._w - _x*q._x - _y*q._y - _z*q._z;
	temp._x = _w*q._x + _x*q._w + _y*q._z - _z*q._y;
	temp._y = _w*q._y - _x*q._z + _y*q._w + _z*q._x;
	temp._z = _w*q._z + _x*q._y - _y*q._x + _z*q._w;
	_w = temp._w; _x = temp._x; _y = temp._y; _z = temp._z;
	return (*this);
}


/**
 * operator / \n
 * Returns the product of the current quaternion q0 and the ref quaternion q.
 * actually q1 = q0/q ==> q1 = q0*q.inverted() 
 * @param &q reference to a quaternion
 * @return reference to this quaternion
 */
inline CQuaternion& CQuaternion::operator/= (const CQuaternion &q){
	(*this) *= q.inverted();
	return (*this);
}


/**
 * operator *= \n
 * Multiplies the elements of the current quaternion by dbl.
 * @param dbl double as a multiplier
 * @return reference to this quaternion
 */
inline CQuaternion& CQuaternion::operator*= (double dbl) {
	_w*=dbl; _x*=dbl; _y*=dbl; _z*=dbl;
	return (*this);
}

/**
 * operator /= \n
 * Divides the elements of the current quaternion by dbl.
 * @param dbl double as a divider
 * @return reference to this quaternion
 */
inline CQuaternion& CQuaternion::operator/= (double dbl) {
	_w/=dbl; _x/=dbl; _y/=dbl; _z/=dbl;
	return (*this);
}

/**
 * operator == \n
 * @param q reference to a quaternion
 * @return true if all elements of ref quaternion are equal 
 * to the corresponding elements of the current quaternion.
 */
inline bool CQuaternion::operator== (const CQuaternion& q) const {
	assert(isValid() && q.isValid());
	return (_w == q._w && _x == q._x && _y == q._y && _z == q._z);
}

/**
 * operator != \n
 * @param q reference to a quaternion
 * @return true if any element of ref quaternion is not equal 
 * to the corresponding element of the current quaternion.
 */
inline bool CQuaternion::operator!= (const CQuaternion& q) const {
	assert(isValid() && q.isValid());
	return (_w != q._w || _x != q._x || _y != q._y || _z != q._z);
}


/**
 * operator < \n
 * This operator is defined so this class can be used in 
 * STL containers.
 * @param q reference to a quaternion
 * @return true if the reference quternion has larger magnitude.
 * however, after normalization, every quaternion has a magnitude of 1
 * @todo find a better metric to compare
 */
inline bool CQuaternion::operator< (const CQuaternion& q) const {
	assert(isValid() && q.isValid());
	return (magnitude() <  q.magnitude());
}


#endif // __QUATERNION_H__
//@}