/** =================================================================\n
 *
 * MathLib\n
 * @file RotationMatrix.h
 *
 * Primary Author: Howard Zhou (howardz@cc.gatech.edu)\n
 *
 * (C) Copyright Howard Zhou 2003\n
 * (C) Copyright Georgia Institute of Technology, 2003\n
 *
 * ==================================================================\n
 * Basic Rotation Matrix Lib
 *
 * @version 1.3\n 
 *     Major organization change, moved most of function defintion to\n
 *     the .cpp file to make class declaration more clear.
 * @date 2003-05-14
 * @author howardz
 * 
 * @version 1.2\n
 *     Implemented getEulerAngle function using tan2
 * @date 2003-04-10
 * @author howardz
 *
 * @version 1.1\n
 *     Added additional constructors for constructing rotation matrix\n 
 *     from 3 vectors indicating a rotation into x, y, z axis
 * @date 2003-03-02
 * @author howardz
 * 
 * @version 1.0\n
 *     Initial Revision
 * @date 2003-02-23
 * @author howardz
 *
 *
 * @todo Revise CRotationMatrix(CVector3 rx, CVector3 ry, CVector3 rz)
 * @bug
 * @warning 
 *
 * ===================================================================\n
 */

#include "RotationMatrix.h"

/// @namespace using standard namespace
using namespace std;


// ======================================================
// Actual function definition
// ======================================================

// ------------------------------------------------------
// constructors and destructors
// ------------------------------------------------------



/**
 * Constructor: Euler angle to rotation
 * Construct a rotation matrix by rotating around 3 principle axis in sequence
 *
 * for rotation of XYZ\n
 * M =\n
 * |                    cos(y)cos(z)                      cos(y)sin(z)         -sin(y) |\n
 * | sin(x)sin(y)cos(z)-cos(x)sin(z)   sin(x)sin(y)sin(z)+cos(x)cos(z)    sin(x)cos(y) |\n
 * | cos(x)sin(y)cos(z)+sin(x)sin(z)   cos(x)sin(y)sin(z)-sin(x)cos(z)    cos(x)cos(y) |\n
 *
 * for rotation of ZXY\n
 * M =\n
 * | cos(y)cos(z)+sin(x)sin(y)sin(z)   cos(x)sin(z)   -sin(y)cos(z)+sin(x)cos(y)sin(z) |\n
 * |-cos(y)sin(z)+sin(x)sin(y)cos(z)   cos(x)cos(z)    sin(y)sin(z)+sin(x)cos(y)cos(z) |\n
 * |                    cos(x)sin(y)        -sin(x)                       cos(x)cos(y) |\n
 *
 * @param xRad rotation angle around x axis in radian
 * @param yRad rotation angle around y axis in radian 
 * @param zRad rotation angle around z axis in radian 
 * @param order the order of rotation
 */
CRotationMatrix::CRotationMatrix (double xRad, double yRad, double zRad, ERotationOrder order) : CMatrix(3,3, 0)
{
	double sx, sy, sz, cx, cy, cz;
	double ss, sc, cc, cs;

	sx = sin(xRad);
	sy = sin(yRad);
	sz = sin(zRad);
	cx = cos(xRad);
	cy = cos(yRad);
	cz = cos(zRad);

	ss = sx*sy;
	sc = sx*cy;
	cc = cx*cy;
	cs = cx*sy;

	// initialize the CMatrix to be the identity CMatrix
	identity();

	switch (order)
	{
	case ZXY:
		// rotateZ(zRad);
		// rotateX(xRad);
		// rotateY(yRad);

		EL(_m, 0, 0) = cy*cz - ss*sz;
		EL(_m, 0, 1) =-cx*sz;
		EL(_m, 0, 2) = sy*cz + sc*sz;
		EL(_m, 1, 0) = cy*sz + ss*cz;
		EL(_m, 1, 1) = cx*cz;
		EL(_m, 1, 2) = sy*sz - sc*cz;
		EL(_m, 2, 0) = -cs;
		EL(_m, 2, 1) = sx;
		EL(_m, 2, 2) = cc;
		break;
	case XYZ:
		// rotateX(xRad);
		// rotateY(yRad);
		// rotateZ(zRad);

		EL(_m, 0, 0) = cy*cz;
		EL(_m, 0, 1) = -cy*sz;
		EL(_m, 0, 2) = sy;
		EL(_m, 1, 0) = ss*cz+cx*sz;
		EL(_m, 1, 1) = -ss*sz+cx*cz;
		EL(_m, 1, 2) = -sc;
		EL(_m, 2, 0) = -cs*cz+sx*sz;
		EL(_m, 2, 1) = cs*sz+sx*cz;
		EL(_m, 2, 2) = cc;

	// using ZXY as default order
	default:
		break;
	}
}


/**
 * Constructor
 * Create matrix that rotates "rx" "ry" and "rz" into the x, y and z axes
 * @param &rx 
 * @param &ry 
 * @param &rz 
 *
 * @todo this constructor will be revised later
 */
CRotationMatrix::CRotationMatrix (const CVector3 &rx, const CVector3 &ry, const CVector3 &rz): CMatrix(3,3, 0)
{
	EL(_m, 0, 0) = rx[0];
	EL(_m, 0, 1) = rx[1];
	EL(_m, 0, 2) = rx[2];
	EL(_m, 1, 0) = ry[0];
	EL(_m, 1, 1) = ry[1];
	EL(_m, 1, 2) = ry[2];
	EL(_m, 2, 0) = rz[0];
	EL(_m, 2, 1) = rz[1];
	EL(_m, 2, 2) = rz[2];
}


// ------------------------------------------------------
// Accessors and Mutators
// ------------------------------------------------------

/**
 * Get the Euler Angle representation of the current rotation matrix
 * @param order the order of rotation, default ZXY
 * @return a set of Euler Angles (rx, ry, rz)
 */
CVector3 CRotationMatrix::getEulerAngles(ERotationOrder order)
{
	double rx, ry, rz;

	switch(order)
	{
	case ZXY:
		rx = asin(EL(_m, 2, 1));
		// if not 90 degree
		ry = atan2(-EL(_m, 2, 0),EL(_m, 2, 2));
		rz = atan2(-EL(_m, 0, 1),EL(_m, 1, 1)); 
		break;
	case XYZ:
		rx = atan2(-EL(_m, 1, 2),EL(_m, 2, 2));
		ry = asin(EL(_m, 0, 2));
		rz = atan2(-EL(_m, 0, 1),EL(_m, 0, 0));
	default:
		break;
	};
	return CVector3(rx, ry, rz);
}

// ------------------------------------------------------
// Output or Test
// ------------------------------------------------------

/**
 * General test function for this class
 * put anything you want to test here.
 */
void CRotationMatrix::test() const
{
	cout << "\n\nCRotationMatrix Test\n=========================\n";

	CRotationMatrix m0(PI/2, PI/2, PI/2);
	CRotationMatrix m1(0, 0, 0);
	CRotationMatrix m2(0, 0, 0);
	CRotationMatrix m3(0, 0, PI/2);

	cout << m0 << endl;
	cout << m1 << endl;
	cout << m2 << endl;
	cout << m3 << endl;

	/*
	m1.rotateZ(PI/2);
	std::cout << m1 << endl;
	m1.rotateX(PI/4);
	std::cout << m1 << endl;
	m1.rotateY(PI/6);
	std::cout << m1 << endl;
	*/

	cout << m2 << endl;
	m2 = CRotationMatrix(PI/4, PI/6, PI/2, ZXY);
	cout << m2 << endl;
}


// rotate around x

void CRotationMatrix::rotateX(double r)
{
	if(0 == r)
		return;

	CRotationMatrix mat;
	double c = cos(r);
	double s = sin(r);

	mat[1][1] = mat[2][2] = c;
	mat[1][2] = -s;
	mat[2][1] = s;

	if(isIdentity())
		(*this) = mat;

	(*this) *= mat;
}

// rotate around y r radian
void CRotationMatrix::rotateY(double r)
{
	if(0 == r)
		return;

	CRotationMatrix mat;
	double c = cos(r);
	double s = sin(r);

	mat[0][0] = mat[2][2] = c;
	mat[0][2] = s;
	mat[2][0] = -s;

	if(isIdentity()) 
		(*this) = mat;

	(*this) *= mat;
}

// rotate around z r radian
void CRotationMatrix::rotateZ(double r)
{
	if(0 == r)
		return;

	CRotationMatrix mat;
	double c = cos(r);
	double s = sin(r);

	mat[0][0] = mat[1][1] = c;
	mat[0][1] = -s;
	mat[1][0] = s;

	if(isIdentity())
		(*this) = mat;
	(*this) *= mat;
}


/*
// Given the x, y, z radian, construct the rotation CMatrix via the given order
CRotationMatrix (double xRad, double yRad, double zRad, ERotationOrder order=XYZ) : CMatrix(3,3, 0)
{
	// initialize the CMatrix to be the identity CMatrix
	identity();

	switch (order)
	{
	case ZXY:
		rotateY(yRad);
		rotateX(xRad);
		rotateZ(zRad);
		break;
	case XYZ:
		rotateZ(zRad);
		rotateY(yRad);
		rotateX(xRad);	
	default:
		break;
		// using XYZ as default order
	}
}
*/


//// forceful constructor
//// TODO: this is dangerous without making sure that 
//// rotationMatrix is orthonormal
//CRotationMatrix (const CMatrix& mat)
//{
//	assert(mat.getRows() == 3 && mat.getCols() == 3);

//	for(int i = 0; i< 3; i++)
//		for(int j=0; j <3; j++)
//			EL(_m, i, j) = mat[i][j];
//}


//@}