/** =================================================================\n
 *
 * MathLib\n
 * @file RotationMatrix.h
 *
 * Primary Author: Howard Zhou (howardz@cc.gatech.edu)\n
 *
 * (C) Copyright Howard Zhou 2003\n
 * (C) Copyright Georgia Institute of Technology, 2003\n
 *
 * ==================================================================\n
 * Basic Rotation Matrix Lib
 *
 * @version 1.3\n 
 *     Major organization change, moved most of function defintion to\n
 *     the .cpp file to make class declaration more clear.
 * @date 2003-05-14
 * @author howardz
 * 
 * @version 1.2\n
 *     Implemented getEulerAngle function using tan2
 * @date 2003-04-10
 * @author howardz
 *
 * @version 1.1\n
 *     Added additional constructors for constructing rotation matrix\n 
 *     from 3 vectors indicating a rotation into x, y, z axis
 * @date 2003-03-02
 * @author howardz
 * 
 * @version 1.0\n
 *     Initial Revision
 * @date 2003-02-23
 * @author howardz
 *
 *
 * @todo Implement det() and invert() 
 * @bug
 * @warning 
 *
 * ===================================================================\n
 */

#ifndef __ROTATIONMATRIX_H__
#define __ROTATIONMATRIX_H__

/// @def PI
#ifndef PI
#define PI 3.14159265358979323846
#endif

#ifndef RADIAN_PER_DEGREE
#define RADIAN_PER_DEGREE 0.0174532925199432957
#define DEGREE_PER_RADIAN 57.295779513082320877
#define RAD(x) ((x)*RADIAN_PER_DEGREE)
#define DEG(x) ((x)*DEGREE_PER_RADIAN)
#endif

#include <iostream>

// local include
#include "Matrix.h"


// fast access to matrix element
#define EL(x,i,j) ((x)[(i)*_cols+(j)])

/// @namespace using standard namespace
using namespace std;


enum ERotationOrder 
{
	XYZ, 
	ZXY
};

// ==================================================================================
/**
 * @class CRotationMatrix RotationMatrix.h RotationMatrix.h
 * @brief Basic Rotational Matrix class (3x3 orthonormal matrix SO3)
 * 
 * @ingroup Lib
 *
 * @par requirements
 *
 * @par license
 * This code is absolutely free to use and modify. The code is provided "as is" with
 * no expressed or implied warranty. The author accepts no liability if it causes
 * any damage to your computer, causes your pet to fall ill, increases baldness
 * or makes your car start emitting strange noises when you start it up.
 * This code has no bugs, just undocumented features!
 * 
 * @version 1.0\n
 *     Initial Revision
 * @date 2003-02-23
 * @author howardz
 *
 * @version 1.1\n
 *     Added additional constructors for constructing rotation matrix\n 
 *     from 3 vectors indicating a rotation into x, y, z axis
 * @date 2003-03-02
 * @author howardz
 * 
 * @version 1.2\n
 *     Implemented getEulerAngle function using tan2, revised\n
 *     matrix from euler angle constructor to be more efficient
 * @date 2003-04-10
 * @author howardz 
 *
 * @version 1.3\n 
 *     Major organization change, moved most of function defintion to\n
 *     the .cpp file to make class declaration more clear.
 * @date 2003-05-14
 * @author howardz
 *
 * @todo 
 * @bug none
 * @warning none
 */
// ==================================================================================
class CRotationMatrix : public CMatrix
{
public:

	// ------------------------------------------------------
	// constructors and destructors
	// ------------------------------------------------------
	CRotationMatrix (double *storage = 0);
	CRotationMatrix (const CRotationMatrix& mat);
	CRotationMatrix (double xRad, double yRad, double zRad, ERotationOrder order = XYZ);
	CRotationMatrix (const CVector3 &rx, const CVector3 &ry, const CVector3 &rz);

	// ------------------------------------------------------
	// Predicators
	// ------------------------------------------------------

	// ------------------------------------------------------
	// Accessors and Mutators
	// ------------------------------------------------------
	CVector3 getEulerAngles   (ERotationOrder order = XYZ);
	CRotationMatrix& invert   (void);
    CRotationMatrix  inverted (void) const; 

	// ------------------------------------------------------
	// Overloaded Operators 
	// ------------------------------------------------------

	// ------------------------------------------------------
	// Output or Test
	// ------------------------------------------------------
	void test() const;

	void rotateX(double r);
	void rotateY(double r);
	void rotateZ(double r);
	// ------------------------------------------------------
	// Friend functions
	// ------------------------------------------------------
};


// ======================================================
// Actual function definition
// ======================================================

// ------------------------------------------------------
// constructors and destructors
// ------------------------------------------------------
/**
 * Default Constructor\n
 * If storage is zero, create a dynamically sized double-precision 
 * CMatrix with the number of rows and cols specified. If storage is 
 * not zero, use the memory it points to.
 * @param *storage memory pointer
 */
inline CRotationMatrix::CRotationMatrix(double *storage) : CMatrix(3, 3, storage)
{
	if(0 == storage)
		identity();
}

/**
 * Copy Constructor\n
 * Creates a CRotationMatrix with the same properties and values as another CRotationMatrix.
 * @param mat reference to another CRotationMatrix
 * @return 
 */
inline CRotationMatrix::CRotationMatrix (const CRotationMatrix& mat) : CMatrix(mat){}


// ------------------------------------------------------
// Accessors and Mutators
// ------------------------------------------------------
/**
 * Inverts the current matrix
 * For rotationMatrix (orthonormal) inversion is just translation
 * @return reference to a matrix
 */
inline CRotationMatrix& CRotationMatrix::invert (void) 
{
	transpose();
	return (*this);
}

/**
 * Returns a matrix that is the inverse of the current matrix.
 * @return a matrix that is the inverse of the current matrix.
 */
inline CRotationMatrix CRotationMatrix::inverted (void) const 
{
	CRotationMatrix mat(*this);
	mat.transpose();
	return CRotationMatrix(mat);
}

#endif // __ROTATIONMATRIX_H__
//@}