/** =================================================================\n
 *
 * MathLib\n
 * @file TransformationMatrix.h
 *
 * Primary Author: Howard Zhou (howardz@cc.gatech.edu)\n
 *
 * (C) Copyright Howard Zhou 2003\n
 * (C) Copyright Georgia Institute of Technology, 2003\n
 *
 * ==================================================================\n
 * Basic Transformation Matrix Lib
 *
 * @version 1.1\n
 *     Added additional constructors for translation and rotation
 *     and transformation
 * @date 2003-03-02
 * @author howardz
 * 
 * @version 1.0\n
 *     Initial Revision
 * @date 2003-02-23
 * @author howardz
 *
 *
 * @todo
 * @bug
 * @warning
 *
 * ===================================================================\n
 */

#include "TransformationMatrix.h"

/// @namespace using standard namespace
using namespace std;

// ======================================================
// Actual function definition
// ======================================================

// ------------------------------------------------------
// constructors and destructors
// ------------------------------------------------------

/**
 * Constructor\n
 * Construct a transformation matrix given 3 rotation angles in radian about 
 * principle axis, and 3 translational distance along principle axis.
 * @param xRad rotation about x axis in radian
 * @param yRad rotation about y axis in radian
 * @param zRad rotation about z axis in radian
 * @param xTrans translation on x axis
 * @param yTrans translation on y axis
 * @param zTrans translation on z axis
 * @param order order of rotation, default XYZ
 */
CTransformationMatrix::CTransformationMatrix (	double xRad, double yRad, double zRad, 
						double xTrans, double yTrans, double zTrans,
						ERotationOrder order) : CMatrix(4, 4, 0)
{
	CRotationMatrix rot(xRad, yRad, zRad);
	CVector3 trans(xTrans, yTrans, zTrans);

	for(int i = 0; i< 3; i++)
		for(int j=0; j <3; j++)
			EL(_m, i, j) = rot[i][j];

	EL(_m, 0, 3) = trans[0];
	EL(_m, 1, 3) = trans[1];
	EL(_m, 2, 3) = trans[2];
	EL(_m, 3, 3) = 1;		
}

/**
 * Transform a CVector3 into another CVector3 (right multiplied by a vector)\n
 * Returns a CVector3 that is the product of the current transformation matrix and refvector.
 * @param refvector reference to a vector
 * @return the resulting vector
 */
CVector3 CTransformationMatrix::operator* (const CVector3& refvector) const 
{	
	CVector4 vec, result;
	vec[0] = refvector[0];
	vec[1] = refvector[1];
	vec[2] = refvector[2];
	vec[3] = 1;

	result[0] = result[1] = result[2] = result[3] = 0;
	for(int i = 0; i < 4; i++)
		for(int j=0; j < 4; j++)
			result[i] += EL(_m, i, j) * vec[j];

	return CVector3(result[0], result[1], result[2]);
}

/**
 * Transformation on vectors in homogeneous form\n
 * homogeneous vector = <x, y, z, w=scale>\n
 * Transform a CVector4 into another CVector4 (right multiplied by a vector)\n
 * Returns a CVector4 that is the product of the current transformation matrix and refvector.
 * @param refvector reference to a CVector4
 * @return the resulting vector
 */
CVector4 CTransformationMatrix::operator* (const CVector4& refvector) const 
{	
	CVector4 vec;

	for(int i = 0; i < 4; i++)
		for(int j=0; j < 4; j++)
			vec[i] += EL(_m, i, j) * refvector[j];

	return CVector4(vec);
}

/**
 * General test function for this class
 * put anything you want to test here.
 */
void CTransformationMatrix::test() const
{
	cout << "\n\nCTransformationMatrix Test\n=========================\n";

	CTransformationMatrix m0(CRotationMatrix( PI/2, PI/2, PI/2), CVector3(1, 2, 3)); 
	CTransformationMatrix m1(PI/2, 0, 0, 1, 2, 3);

	cout << m0 << endl;
	cout << m1 << endl;
}

//// forceful constructor
//// TODO: this is dangerous without making sure that 
//// rotationMatrix is orthonormal
//CTransformationMatrix (const CMatrix &mat): CMatrix(4, 4, 0)
//{
//	assert(mat.getRows() == 4 && mat.getCols() == 4);
//	for(int i = 0; i< 4; i++)
//		for(int j=0; j <4; j++)
//			EL(_m, i, j) = mat[i][j];
//}

//@}
