/** =================================================================\n
 *
 * MathLib\n
 * @file TransformationMatrix.h
 *
 * Primary Author: Howard Zhou (howardz@cc.gatech.edu)\n
 *
 * (C) Copyright Howard Zhou 2003\n
 * (C) Copyright Georgia Institute of Technology, 2003\n
 *
 * ==================================================================\n
 * Basic Transformation Matrix Lib
 *
 * @version 1.1\n
 *     Added additional constructors for translation and rotation
 *     and transformation
 * @date 2003-03-02
 * @author howardz
 * 
 * @version 1.0\n
 *     Initial Revision
 * @date 2003-02-23
 * @author howardz
 *
 *
 * @todo Implement det() and invert() 
 * @bug
 * @warning 
 *
 * ===================================================================\n
 */

#ifndef __TRANSFORMATIONMATRIX_H__
#define __TRANSFORMATIONMATRIX_H__


// local include
#include "Matrix.h"
#include "RotationMatrix.h"


// fast access to matrix element
#define EL(x,i,j) ((x)[(i)*_cols+(j)])
// using namespace std;

// ==================================================================================
/**
 * @class CTransformationMatrix TransformationMatrix.h TransformationMatrix.h
 * @brief Basic Transformation Matrix Lib (4x4)\n
 * with a 3x3 rotational matrix on its upper left corner, 3x1 translational vector
 * on its upper right corner, 1x3 scale vector on its lower left corner and 
 * 1 on the lower right corner\n
 * | r     r     r     t |\n
 * | r     r     r     t |\n
 * | r     r     r     t |\n
 * | s     s     s     1 |\n
 * 
 * @ingroup Lib
 *
 * @par requirements
 *
 * @par license
 * This code is absolutely free to use and modify. The code is provided "as is" with
 * no expressed or implied warranty. The author accepts no liability if it causes
 * any damage to your computer, causes your pet to fall ill, increases baldness
 * or makes your car start emitting strange noises when you start it up.
 * This code has no bugs, just undocumented features!
 * 
 *
 * @version 1.1\n
 *     Added additional constructors for translation and rotation
 *     and transformation
 * @date 2003-03-02
 * @author howardz
 * 
 * @version 1.0\n
 *     Initial Revision
 * @date 2003-02-23
 * @author howardz
 *
 *
 * @todo 
 * @bug
 * @warning
 */
// ==================================================================================

class CTransformationMatrix : public CMatrix
{
public:

	// ------------------------------------------------------
	// constructors and destructors
	// ------------------------------------------------------
	CTransformationMatrix (double *storage = 0);
	CTransformationMatrix (const CTransformationMatrix &mat);
	CTransformationMatrix (const CRotationMatrix &rot, const CVector3 &trans);
	CTransformationMatrix (const CVector3 &trans);
	CTransformationMatrix (const CRotationMatrix &rot);
	CTransformationMatrix (	double xRad, double yRad, double zRad, 
							double xTrans, double yTrans, double zTrans,
							ERotationOrder order = XYZ);

	// ------------------------------------------------------
	// Predicators
	// ------------------------------------------------------

	// ------------------------------------------------------
	// Accessors and Mutators
	// ------------------------------------------------------
	CRotationMatrix getRotation(void) const;
	CVector3 getTranslation(void) const;

	// ------------------------------------------------------
	// Overloaded Operators 
	// ------------------------------------------------------
	CVector3 operator* (const CVector3& refvector) const;
	CVector4 operator* (const CVector4& refvector) const;

	// ------------------------------------------------------
	// Output or Test
	// ------------------------------------------------------
	void test() const;

	// ------------------------------------------------------
	// Friend functions
	// ------------------------------------------------------
};

// ======================================================
// Actual function definition
// ======================================================

// ------------------------------------------------------
// constructors and destructors
// ------------------------------------------------------

/**
 * Default Constructor\n
 * If storage is zero, create a dynamically sized double-precision 
 * CMatrix with the number of rows and cols specified. If storage is 
 * not zero, use the memory it points to.
 * @param *storage pointer to previously allocated storage 
 */
inline CTransformationMatrix::CTransformationMatrix(double *storage) : CMatrix(4, 4, storage)
{
	if(0 == storage)
		identity();
}

/**
 * Copy Constuctor\n
 * Creates a CMatrix with the same properties and values as another CMatrix.
 * @param &mat reference to a matrix
 */
inline CTransformationMatrix::CTransformationMatrix (const CTransformationMatrix &mat) : CMatrix(mat){}

/**
 * Constructor\n
 * Constructs a transformation matrix given a roational matrix and a translational vector
 * @param &rot reference to a rotational matrix
 * @param &trans reference to a translational vector
 */
inline CTransformationMatrix::CTransformationMatrix (const CRotationMatrix &rot, const CVector3 &trans):CMatrix(4, 4, 0)
{
	for(int i = 0; i< 3; i++)
		for(int j=0; j <3; j++)
			EL(_m, i, j) = rot[i][j];

	EL(_m, 0, 3) = trans[0];
	EL(_m, 1, 3) = trans[1];
	EL(_m, 2, 3) = trans[2];
	EL(_m, 3, 3) = 1;
}

/**
 * Constructor\n
 * Constructs a transformation matrix given a translational vector
 * @param &trans reference to a translational vector
 */
inline CTransformationMatrix::CTransformationMatrix (const CVector3 &trans):CMatrix(4, 4, 0)
{
	identity();
	EL(_m, 0, 3) = trans[0];
	EL(_m, 1, 3) = trans[1];
	EL(_m, 2, 3) = trans[2];
}	

/**
 * Constructor\n
 * Constructs a transformation matrix given a rotational matrix
 * @param &rot reference to a rotational matrix
 */
inline CTransformationMatrix::CTransformationMatrix (const CRotationMatrix &rot):CMatrix(4, 4, 0)
{
	for(int i = 0; i< 3; i++)
		for(int j=0; j <3; j++)
			EL(_m, i, j) = rot[i][j];

	EL(_m, 3, 3) = 1;
}


// ------------------------------------------------------
// Accessors and Mutators
// ------------------------------------------------------

/**
 * Returns the rotational part of the current transformation
 * @return the rotational matrix embedded in the current transformation matrix
 */
inline CRotationMatrix CTransformationMatrix::getRotation(void) const
{
	CRotationMatrix rot;

	for(int i = 0; i< 3; i++)
		for(int j=0; j <3; j++)
			rot[i][j] = EL(_m, i, j);

	return CRotationMatrix(rot);
}

/**
 * Returns the translational part of the current transformation
 * @return the translational vector embedded in the current transformation matrix
 */
inline CVector3 CTransformationMatrix::getTranslation(void) const
{
	return CVector3(EL(_m, 0,3), EL(_m,1,3), EL(_m,2,3));
}

#endif // __TRANSFORMATIONMATRIX_H__
