/** =================================================================\n
 *
 * CS 7491 2004\n
 * @file PointSet.h
 *
 * Primary Authors: Brandon Beck
 *					Nguyen Truong (nltruong@cc.gatech.edu)\n
 *                  Howard Zhou (howardz@cc.gatech.edu)\n
 *
 * (C) Copyright 2004\n
 * (C) Copyright Georgia Institute of Technology, 2004\n
 *
 * ==================================================================\n
 * Point set generation and operation
 *
 * @version 1.0\n
 *     Initial Revision
 * @date 2004-02-26
 * @author Brandon Beck, Nguyen Truong, Howard Zhou
 *
 * @todo
 * @bug
 * @warning 
 *
 * ===================================================================\n
 */

#ifndef __CPOINTSET_H__
#define __CPOINTSET_H__

#include <vector>
#include <time.h>
#include "..\MatrixLib\Vector.h"

#ifndef GEOMETRY_TYPE
#define UNKNOWN 0
#define SPHERE	1
#define BALL	2
#define BLOCKER 3
#endif


#ifndef PI
#define PI 3.141592654
#endif

/// @namespace using standard namespace
using namespace std;


/// local function declaration

CPoint3D pointOnSphere	(CPoint3D center, double radius);
CPoint3D pointInBall	(CPoint3D center, double radius);
vector<CPoint3D> pointsOnSphere	(long nPoints, CPoint3D center, double radius, bool bRelax = false);
vector<CPoint3D> pointsInBall	(long nPoints, CPoint3D center, double radius, bool bRelax = false);

void relaxPointsOnSphere(vector<CPoint3D>& points, 
						 CPoint3D sphereCenter, double sphereRadius, 
						 double radiusOfInfluence, double k = 0.2);
void relaxPointsInBall(vector<CPoint3D>& points, 
					   CPoint3D ballCenter, double ballRadius, 
					   double radiusOfInfluence, double k = 0.2);

/**
 * class geometry
 **/
template <class T>
class CGeometry
{
protected:

public:
	int		_type;
	T		_center;
	double	_radius;

	// ------------------------------------------------------
	// constructors and destructors
	// ------------------------------------------------------

	/// Default constructor. Creates an uninitialized geometry
	CGeometry(){};

	CGeometry(int type, T center, double radius):
		_type(type),
		_center(center),
		_radius(radius){}
	/*
	{
		_type	= type;
		_center	= center;
		_radius	= radius;
	}
	*/

	// copy constructor	
	CGeometry(const CGeometry<T>& refgeo):
		_type(refgeo._type),
		_center(refgeo._center),
		_radius(refgeo._radius)
		{}

	// Destructor
	~CGeometry(){};

	// ------------------------------------------------------
	// Predicators
	// ------------------------------------------------------
	virtual bool isValid() const {return true;};
	virtual bool contains(T p) const {return true;};
};

/**
 * template specialization
 **/
inline
bool CGeometry<CPoint3D>::contains(CPoint3D p) const
{
	CVector3 d = p - _center;
	if(d.magnitude() <= _radius)
		return true;
	return false;
}


// ==================================================================================
/**
 * @class CPointSet
 * @brief PointSet class definition
 * @ingroup Set
 *
 * @par requirements
 *
 * @par license
 * This code is absolutely free to use and modify. The code is provided "as is" with
 * no expressed or implied warranty. The author accepts no liability if it causes
 * any damage to your computer, causes your pet to fall ill, increases baldness
 * or makes your car start emitting strange noises when you start it up.
 * This code has no bugs, just undocumented features!
 *
 * 
 * @version 1.0\n
 *     Initial Revision
 * @date 2004-02-26
 * @author Brandon Beck, Nguyen Truong & Howard Zhou
 *
 * This class is a general point set class. This is an abstract class serving as the
 * point set interface.
 * 
 * @todo
 * @bug
 * @warning
 */
// ================================================================================

template <class T>
class CPointSet
{
protected:
	vector<T> _points;		///< place to store the points in this set 	
	// bool	  _bFill;		///< make it a volumn or surface
	CGeometry<T> _geometry;	///< boundary geometry of this point set

private:
	// ------------------------------------------------------
	// private helpers
	// ------------------------------------------------------

public:
	// ------------------------------------------------------
	// constructors and destructors
	// ------------------------------------------------------

	///< Default constructor. Creates an uninitialized pointset.
	CPointSet(){};

	///< copy constructor
	CPointSet(const CPointSet& refset):
		_points(refset._points),
		_geometry(refset._geometry){}

	///< generate uniformly distributed points according to the given boundary type
	CPointSet(long nPoints, CGeometry<CPoint3D> geometry, bool bRelax = false);
	CPointSet(long nPoints, int type, CPoint3D center, double radius, bool bRelax = false);

	CPointSet(double avgDist, int type, CPoint3D center, double radius, bool bRelax = false);

	// Destructor
	~CPointSet(){};

	// ------------------------------------------------------
	// Predicators
	// ------------------------------------------------------
	virtual bool isValid() const {return true;};
	virtual bool contains(const CPointSet<T>& refset ) const;

	// ------------------------------------------------------
	// Accessors and Mutators
	// ------------------------------------------------------
	vector<T> getPoints		(void) const {return _points;};
	long	  getPointCount	(void) const {return (long)_points.size();};

	// ------------------------------------------------------
	// Overloaded Operators 
	// ------------------------------------------------------
	CPointSet<T>	operator+  (const	CPointSet<T>& refset) const;
	CPointSet<T>	operator-  (const	CPointSet<T>& refset) const;
	CPointSet<T>&	operator+= (const	CPointSet<T>& refset);
	CPointSet<T>&	operator-= (const	CPointSet<T>& refset);
	CPointSet<T>&	operator = (const	CPointSet<T>& refset);
	//bool		operator== (const	CPointSet& refset) const;
	//bool		operator!= (const	CPointSet& refset) const;
	//bool		operator<  (const	CPointSet& refset) const;


	// ------------------------------------------------------
	// Input, Output, and Test
	// ------------------------------------------------------
	virtual void readFile	(const char* filename);
	virtual void writeFile	(const char* filename) const;

	// ------------------------------------------------------
	// Other operation
	// ------------------------------------------------------
	double getAverageDistance(int nNeighbours) const;

	// ------------------------------------------------------
	// Friend functions
	// ------------------------------------------------------

};

// NOTE:
// For a template class, all the function definition has to 
// be inside the header file


// ------------------------------------------------------
// constructors and destructors
// ------------------------------------------------------

/**
 * generate uniformly distributed points according to the given boundary type
 *
 **/
CPointSet<CPoint3D>::CPointSet(long nPoints, CGeometry<CPoint3D> geometry, bool bRelax)
{
	switch(geometry._type)
	{
	case BALL:
		_points = pointsInBall(nPoints, geometry._center, geometry._radius, bRelax);
		_geometry = geometry;
		break;
	case SPHERE:
		_points = pointsOnSphere(nPoints, geometry._center, geometry._radius, bRelax);
		_geometry = geometry;
		break;
	case BLOCKER:
		_geometry = geometry;
		break;
	default:
		_geometry = geometry;
	}
}

/**
 * generate uniformly distributed points according to the given boundary type
 *
 **/
CPointSet<CPoint3D>::CPointSet(long nPoints, int type, CPoint3D center, double radius, bool bRelax)
{
	switch(type)
	{
	case BALL:
		_points = pointsInBall(nPoints, center, radius, bRelax);
		_geometry = CGeometry<CPoint3D>(BALL, center, radius);
		break;
	case SPHERE:
		_points = pointsOnSphere(nPoints, center, radius, bRelax);
		_geometry = CGeometry<CPoint3D>(SPHERE, center, radius);
		break;
	case BLOCKER:
		_geometry = CGeometry<CPoint3D>(BALL, center, radius);
		break;
	default:
		_geometry = CGeometry<CPoint3D>(BALL, center, radius);
	}
}

CPointSet<CPoint3D>::CPointSet(double avgDist, int type, CPoint3D center, double radius, bool bRelax)
{
	double volume, area;
	long nPoints = 0;
	switch(type)
	{
	case BALL:
		// calculate the number of points we need for the ball
		volume = 4.0 * PI * radius * radius * radius / 3.0;
		nPoints = volume / (avgDist * avgDist * avgDist);
		_points = pointsInBall(nPoints, center, radius, bRelax);
		_geometry = CGeometry<CPoint3D>(BALL, center, radius);
		break;
	case SPHERE:
		// calculate the number of points we need for the sphere
		area = 4.0 * PI * radius * radius;
		nPoints = area / (avgDist * avgDist);
		_points = pointsOnSphere(nPoints, center, radius, bRelax);
		_geometry = CGeometry<CPoint3D>(SPHERE, center, radius);
		break;
	case BLOCKER:
		_geometry = CGeometry<CPoint3D>(BALL, center, radius);
		break;
	default:
		_geometry = CGeometry<CPoint3D>(BALL, center, radius);
	}
	cout << "# of points generated : " << nPoints << endl;

}


// ------------------------------------------------------
// Other operation
// ------------------------------------------------------
/**
 * relax, relaxed
 * 
 * though uniform density, when the number of points is
 * small, the distribution would be random and fairly
 * nonuniform. Therefore, we relax points by giving them
 * a repulsion from the neighbouring points
 **/

/**
 * getAverageDistance
 * get the average distance to its nearest nNeighbours points
 *
 **/
double CPointSet<CPoint3D>::getAverageDistance(int nNeighbours) const
{
	return 0;	
}

// ------------------------------------------------------
// Predicators
// ------------------------------------------------------
template <class T>
inline bool CPointSet<T>::contains(const CPointSet<T>& refset) const
{
	return false;
}

// ------------------------------------------------------
// Input, Output, and Test
// ------------------------------------------------------
void CPointSet<CPoint3D>::readFile(const char* filename)
{
	// clear what's left
	_points.clear();

	FILE * inFile;
	char line[255];
	long i, nV;
	double vx, vy, vz;

	// open the file
	if(NULL == (inFile = fopen(filename, "rb")))
	{
		cout << "Error -- unable to open file <" << filename << ">\n";
		exit(-1);
	}

	// get the vertex count
	if(NULL == fgets(line, 255, inFile) || (nV = atol(line)) == 0)
	{
		cout << "Error -- wrong file format\n";
		exit(-1);
	}

	// get vertex coordinates
	for(i = 0; i < nV; i++)
	{
		fscanf(inFile, "%lf %lf %lf\n", &vx, &vy, &vz);
		_points.push_back(CPoint3D(vx, vy, vz));
	}

	fclose(inFile);
}



void CPointSet<CPoint3D>::writeFile(const char* filename) const
{
	long i, nV;
	FILE * outFile;
	CPoint3D p;

	if(NULL == (outFile = fopen(filename, "w")))
	{
		cout << "Error -- unable to open file <" << filename << ">\n";
		exit(-1);
	}

	nV = long(_points.size());


	fprintf(outFile, "%ld\n", nV);

	for(i = 0; i < nV; i++)
	{
		p = _points[i];
		fprintf(outFile, "%f %f %f\n", float(p[0]) ,float(p[1]), float(p[2]));
	}

	fclose(outFile);
}

// ------------------------------------------------------
// Overloaded Operators 
// ------------------------------------------------------

/**
 * assignment operator
 **/
template <class T>
inline CPointSet<T>& CPointSet<T>::operator = (const CPointSet<T>& refset)
{
	_points		= refset._points;
	_geometry	= refset._geometry;

	return (*this);
}

/**
 * operator +
 *
 * take the union of both point set
 *
 **/
template <class T>
inline CPointSet<T> CPointSet<T>::operator + (const CPointSet<T>& refset) const
{
	CPointSet<T> temp;
	temp = *this;
	temp._points.insert(temp._points.end(), refset._points.begin(), refset._points.end());
	return temp;
}

/**
 * operator -
 * take the difference of the two point set. That is, if any point in
 * the first set is somehow inside the second set (or the range of the
 * second set), then disrecard them
 **/
template<class T>
CPointSet<T> CPointSet<T>::operator - (const CPointSet<T>& refset) const
{
	CPointSet temp;
	// temp = *this;

	vector<T>::const_iterator curr;

	for(curr = _points.begin();
		curr != _points.end();
		curr ++)
	{
		if(!refset._geometry.contains(*curr))
		{
			temp._points.push_back(*curr);
		}
	}
	return temp;
}

/**
 * operator +=
 *
 * take the union of both point set
 *
 **/
template<class T>
inline CPointSet<T>& CPointSet<T>::operator += (const CPointSet<T>& refset)
{
	_points.insert(_points.end(), refset._points.begin(), refset._points.end());
	return (*this);
}

/**
 * operator -=
 * take the difference of the two point set. That is, if any point in
 * the first set is somehow inside the second set (or the range of the
 * second set), then disrecard them
 **/
template<class T>
CPointSet<T>& CPointSet<T>::operator -= (const CPointSet<T>& refset)
{
	CPointSet temp;
	vector<T>::iterator curr;

	for(curr = _points.begin();
		curr = _points.end();
		curr ++)
	{
		if(!refset._geometry.contains(*curr))
		{
			temp._points.push_back(*curr);
		}
	}
	_points = temp._points;
	return (*this);
}




/**
 * relaxation of points on a shpere
 * O(n^2) algorithm
 *
 * @param radius of influence, outside this radius would be considered
 *		  to have no influence over the current point
 * @param k small constant
 **/
void relaxPointsInBall(vector<CPoint3D>& points, 
					   CPoint3D ballCenter, double ballRadius, 
					   double radiusOfInfluence, double k)
{
	unsigned int i, j;

	CVector3 qp, repulsion, repulsionSum;
	CPoint3D p, p1, p2, q;

	for(i = 0; i < points.size(); i++)
	{
		repulsionSum = CVector3(0, 0, 0);
		p = points[i];
		for(j = 0; j < points.size(); j++)
		{
			q = points[j];

			// if q is the same point or q is located 
			// too far away from p, discard the point
			if((p-q).magnitude() > radiusOfInfluence || p == q)
				continue;
			
			qp = p - q;
			// the repulsion force is inversely related to the
			// distance.
			repulsion = (p-q).normalized() * (radiusOfInfluence/qp.magnitude());
			repulsionSum += repulsion;
		}

		// new position not limited to be inside the ball
		p1 = p + repulsionSum * k;

		// if p1 goes outside the ball, put it back into the ball
		if((p1 - ballCenter).magnitude() > ballRadius) 
		{
			p2 = ballCenter + (p1 - ballCenter).normalized() * ballRadius;
			points[i] = p2;
		}
		else
		{
			points[i] = p1;
		}		
	}
}


/**
 * relaxation of points on a shpere
 * O(n^2) algorithm
 *
 * @param radius of influence, outside this radius would be considered
 *		  to have no influence over the current point
 * @param k small constant
 **/

void relaxPointsOnSphere(vector<CPoint3D>& points, 
						 CPoint3D sphereCenter, double sphereRadius, 
						 double radiusOfInfluence, double k)
{
	unsigned int i, j;

	CVector3 qp, repulsion, repulsionSum;
	CPoint3D p, p1, p2, q;

	for(i = 0; i < points.size(); i++)
	{
		repulsionSum = CVector3(0, 0, 0);
		p = points[i];
		for(j = 0; j < points.size(); j++)
		{
			q = points[j];

			// if q is the same point or q is located 
			// too far away from p, discard the point
			if((p-q).magnitude() > radiusOfInfluence || p == q)
				continue;
			
			qp = p - q;
			// the repulsion force is inversely related to the
			// distance.
			repulsion = (p-q).normalized() * (radiusOfInfluence/qp.magnitude());
			repulsionSum += repulsion;
		}

		// new position not on the sphere
		p1 = p + repulsionSum * k;

		// put p1 back onto the sphere;
		p2 = sphereCenter + (p1 - sphereCenter).normalized() * sphereRadius;
		points[i] = p2;
	}
}



/* Returns a random point on the sphere with center and radius 
 * Must include stdlib.h for RAND_MAX 
 */
CPoint3D pointOnSphere(CPoint3D center, double radius)
{	
	double x0 = center[0];
	double y0 = center[1];
	double z0 = center[2];
	
	double slice = (((double)rand()/RAND_MAX) - 0.5) * 2 * radius;
	double angle = ((double)rand()/RAND_MAX) * 2 * PI;
	double new_radius = sqrt(radius*radius - slice*slice);
	double new_x = slice + x0;	
	double new_y = new_radius * cos(angle) + y0;
	double new_z = new_radius * sin(angle) + z0;

	return CPoint3D(new_x, new_y, new_z);
}


/**
 * pointsOnSphere
 *
 * produce points lies on a sphere with blocker
 * 
 * @param nPoints number of points to generate on the given sphere; however, if there is a blocker present
 *        to maintain the same uniform density, we will just remove points already generated, so the
 *        final number of points may be less than this number.
 * @param radiusBlocker the radius of the blocker, if this is set to 0, then there is no blocker
 *
 **/
vector<CPoint3D> pointsOnSphere(long nPoints, CPoint3D centerSphere, double radiusSphere, bool bRelax)
{
	long i;
	vector<CPoint3D> points;

	// Seed the random-number generator with current time so that
	// the numbers will be different every time we run.
	srand( (unsigned)time( NULL ) );

	// generate the points
	for(i = 0; i < nPoints; i++)
	{
		CPoint3D p = pointOnSphere(centerSphere, radiusSphere);
		points.push_back(p);
	}

	if (bRelax)
	{
		// for relaxation of the points on the sphere
		double sphereArea = 4 * PI * radiusSphere * radiusSphere; 
		double radiusOfInfluence = 2 * sqrt(sphereArea / nPoints);
		double k = 0.2;
		// relax the points
		for(i = 0; i< 10; i++)
		{
			relaxPointsOnSphere(points, centerSphere, radiusSphere, radiusOfInfluence, k);
			k -= 0.015;
		}
	}

	return points;
}



/* Returns a random point in the ball with center and radius
 * Must include stdlib.h for RAND_MAX
 */
CPoint3D pointInBall(CPoint3D center, double radius) 
{
	double x, y, z;
	// random number with [-1, 1]
	do
	{
		x = (((double)rand()/RAND_MAX) - 0.5) * 2;
		y = (((double)rand()/RAND_MAX) - 0.5) * 2;
		z = (((double)rand()/RAND_MAX) - 0.5) * 2;
	}while (x*x + y*y + z*z >= 1);

	CPoint3D pStd = CPoint3D(x, y, z);
	
	return CPoint3D(pStd * radius + center);
}



/**
 * pointsInBall
 *
 * produce points lies on a ball with blocker
 * 
 * @param nPoints number of points to generate on the given sphere; however, if there is a blocker present
 *        to maintain the same uniform density, we will just remove points already generated, so the
 *        final number of points may be less than this number.
 * @param radiusBlocker the radius of the blocker, if this is set to 0, then there is no blocker
 *
 **/
vector<CPoint3D> pointsInBall(long nPoints, CPoint3D centerBall, double radiusBall, bool bRelax)
{
	long i;
	vector<CPoint3D> points;


	// Seed the random-number generator with current time so that
	// the numbers will be different every time we run.
	srand( (unsigned)time( NULL ) );

	// generate the points
	for(i = 0; i < nPoints; i++)
	{
		CPoint3D p = pointInBall(centerBall, radiusBall);
		points.push_back(p);
	}

	if (bRelax)
	{
		double ballVolume = 4 / 3 * PI * radiusBall * radiusBall * radiusBall; 
		double radiusOfInfluence = 2 * pow((ballVolume / nPoints), 1/3);
		// relax the points
		double k = 0.07;
		for(i = 0; i< 10; i++)
		{
			relaxPointsInBall(points, centerBall, radiusBall, radiusOfInfluence, k);
			k -= 0.005;
		}
	}

	return points;
	// remove the points inside the blocker
	/*
	if(radiusBlocker == 0)
	{	
		return points;
	}
	else
	{
		vector<CPoint3D> temp;

		for(i = 0; i < nPoints; i++)
			if (!isIn(points[i], centerBlocker, radiusBlocker))
				temp.push_back(points[i]);
		return temp;
	}
	*/
}


#endif // __CPOINTSET_H__